"""A set of functions for checking an environment's details.

Based on module in the Gymnasium repository:
https://github.com/Farama-Foundation/Gymnasium/blob/v0.27.0/gymnasium/utils/env_checker.py

Which in turn is based on work from:

- Stable Baselines3 repository hosted on GitHub
  (https://github.com/DLR-RM/stable-baselines3/). Original Author: Antonin Raffin.
- warnings/assertions from the PettingZoo repository hosted on GitHub
  (https://github.com/PettingZoo-Team/PettingZoo). Original Author: J K Terry/.

These projects are covered by the MIT License.

"""
import inspect
from copy import deepcopy

import posggym
from posggym import error, logger
from posggym.utils.passive_env_checker import (
    check_rng_equality,
    data_equivalence,
    check_agent_action_spaces,
    check_agent_observation_spaces,
    check_agent_space_limits,
    check_reset_obs,
    env_render_passive_checker,
    env_reset_passive_checker,
    env_step_passive_checker,
)


def check_reset_seed(env: posggym.Env):
    """Check that the environment can be reset with a seed.

    Arguments
    ---------
    env: The environment to check

    Raises
    ------
    AssertionError: The environment cannot be reset with a random seed,
        even though `seed` or `kwargs` appear in the signature.

    """
    signature = inspect.signature(env.reset)
    if "seed" in signature.parameters or (
        "kwargs" in signature.parameters
        and signature.parameters["kwargs"].kind is inspect.Parameter.VAR_KEYWORD
    ):
        try:
            obs_1, info = env.reset(seed=123)
            check_reset_obs(obs_1, env.model)

            assert (
                env.unwrapped.model._rng  # pyright: ignore [reportPrivateUsage]
                is not None
            ), (
                "Expects the random number generator to have been generated given a "
                "seed was passed to reset. Mostly likely the environment reset function"
                " does not call `super().reset(seed=seed)`."
            )
            seed_123_rng = deepcopy(
                env.unwrapped.model._rng  # pyright: ignore [reportPrivateUsage]
            )

            obs_2, info = env.reset(seed=123)
            check_reset_obs(obs_2, env.model)

            if env.spec is not None and env.spec.nondeterministic is False:
                assert data_equivalence(obs_1, obs_2), (
                    "Using `env.reset(seed=123)` is non-deterministic as the "
                    "observations are not equivalent."
                )
            check_rng_equality(
                env.unwrapped.model._rng,  # pyright: ignore [reportPrivateUsage]
                seed_123_rng,
                prefix=(
                    "Mostly likely the environment reset function does not call "
                    "`super().reset(seed=seed)` as the random generators are not same "
                    "when the same seeds are passed to `env.reset`. "
                ),
            )

            obs_3, info = env.reset(seed=456)
            check_reset_obs(obs_3, env.model)
            try:
                check_rng_equality(
                    env.unwrapped.model._rng,  # pyright: ignore [reportPrivateUsage]
                    seed_123_rng,
                    prefix="",
                )
            except AssertionError:
                pass
            else:
                raise AssertionError(
                    "Mostly likely the environment reset function does not call "
                    "`super().reset(seed=seed)` as the random number generators are not"
                    " different when different seeds are passed to `env.reset`."
                )

        except TypeError as e:
            raise AssertionError(
                "The environment cannot be reset with a random seed, even though "
                "`seed` or `kwargs` appear in the signature. This should never happen, "
                f"please report this issue. The error was: {e}"
            ) from e

        seed_param = signature.parameters.get("seed")
        # Check the default value is None
        if seed_param is not None and seed_param.default is not None:
            logger.warn(
                "The default seed argument in reset should be `None`, otherwise the "
                "environment will by default always be deterministic. "
                f"Actual default: {seed_param.default}"
            )
    else:
        raise error.Error(
            "The `reset` method does not provide a `seed` or `**kwargs` keyword "
            "argument."
        )


def check_reset_options(env: posggym.Env):
    """Check that the environment can be reset with options.

    Arguments
    ---------
    env: The environment to check

    Raises
    ------
    AssertionError: The environment cannot be reset with options, even though
        `options` or `kwargs` appear in the signature.

    """
    signature = inspect.signature(env.reset)
    if "options" in signature.parameters or (
        "kwargs" in signature.parameters
        and signature.parameters["kwargs"].kind is inspect.Parameter.VAR_KEYWORD
    ):
        try:
            env.reset(options={})
        except TypeError as e:
            raise AssertionError(
                "The environment cannot be reset with options, even though `options` "
                "or `**kwargs` appear in the signature. This should never happen, "
                "please report this issue. The error was: {e}"
            ) from e
    else:
        raise error.Error(
            "The `reset` method does not provide an `options` or `**kwargs` keyword "
            "argument."
        )


def check_reset_return_type(env: posggym.Env):
    """Checks that :meth:`reset` correctly returns a tuple of the form `(obs , info)`.

    Arguments
    ---------
    env: The environment to check

    Raises
    ------
    AssertionError depending on spec violation

    """
    result = env.reset()
    assert isinstance(result, tuple), (
        "The result returned by `env.reset()` was not a tuple of the form "
        "`(obs, info)`, where `obs` is a observation and `info` is a dictionary "
        f"containing additional information. Actual type: `{type(result)}`"
    )
    assert (
        len(result) == 2
    ), f"Calling reset method did not return a 2-tuple, actual length: {len(result)}"

    obs, info = result
    check_reset_obs(obs, env.model)
    assert isinstance(info, dict), (
        "The second element returned by `env.reset()` was not a dictionary, "
        f"actual type: {type(info)}"
    )


def check_env(env: posggym.Env, skip_render_check: bool = False):
    """Check that an environment follows posggym API.

    This is an invasive function that calls the environment's reset and step.

    This is particularly useful when using a custom environment.

    TODO Update links

    Please take a look at https://gymnasium.farama.org/content/environment_creation/
    for more information about the API.

    Arguments
    ---------
    env: The posggym environment that will be checked
    skip_render_check: Whether to skip the checks for the render method.
        True by default (useful for the CI)

    """
    more_info_msg = "See COMING SOON for more info."

    assert isinstance(
        env, posggym.Env
    ), f"The environment must inherit from the posggym.Env class. {more_info_msg}"

    if env.unwrapped is not env:
        logger.warn(
            f"The environment ({env}) is different from the unwrapped version "
            f"({env.unwrapped}). This could effect the environment checker as the "
            "environment most likely has a wrapper applied to it. We recommend using "
            "the raw environment for `check_env` using `env.unwrapped`."
        )

    # ============= Check the spaces (observation and action) ================
    assert hasattr(
        env, "action_spaces"
    ), f"The environment must specify agent action spaces. {more_info_msg}"
    check_agent_action_spaces(env.action_spaces)
    check_agent_space_limits(env.action_spaces, "action")

    assert hasattr(
        env, "observation_spaces"
    ), f"The environment must specify agent observation spaces. {more_info_msg}"
    check_agent_observation_spaces(env.observation_spaces)
    check_agent_space_limits(env.observation_spaces, "observation")

    # ==== Check the reset method ====
    check_reset_return_type(env)
    check_reset_seed(env)
    check_reset_options(env)

    # ============ Check the returned values ===============
    env_reset_passive_checker(env)
    env_step_passive_checker(
        env, {i: env.action_spaces[i].sample() for i in env.agents}
    )

    # ==== Check the render method and the declared render modes ====
    if not skip_render_check:
        if env.render_mode is not None:
            env_render_passive_checker(env)

        for render_mode in env.metadata["render_modes"]:
            assert env.spec is not None
            new_env = env.spec.make(render_mode=render_mode)
            new_env.reset()
            env_render_passive_checker(new_env)
            new_env.close()



"""Wrapper for limiting the time steps of an environment."""
from typing import Optional, Set

import posggym


class TimeLimit(posggym.Wrapper):
    """Wraps environment to enforce environment time limit.

    This wrapper will issue a `truncated` signal in the :meth:`step` method for any
    agents that have not already reached a terminal state by the time a maximum number
    of timesteps is exceeded. It will also signal that the episode is `done` for all
    agents.

    Arguments
    ---------
    env : posggym.Env
      The environment to apply the wrapper
    max_episode_steps : int, optional
      The maximum length of episode before it is truncated. If None then will not
      truncate episodes.

    Note
    ----
    This implementation is based on the similar Gymnasium wrapper:
    https://github.com/Farama-Foundation/Gymnasium/blob/v0.27.0/gymnasium/wrappers/time_limit.py

    """

    def __init__(self, env: posggym.Env, max_episode_steps: Optional[int] = None):
        super().__init__(env)
        if max_episode_steps is None and self.env.spec is not None:
            assert env.spec is not None
            max_episode_steps = env.spec.max_episode_steps
        if self.env.spec is not None:
            self.env.spec.max_episode_steps = max_episode_steps
        self._max_episode_steps = max_episode_steps
        self._elapsed_steps = 0
        self._terminated_agents: Set[str] = set()

    def step(self, actions):
        obs, rewards, terminated, truncated, done, info = self.env.step(actions)
        self._elapsed_steps += 1

        if self._elapsed_steps >= self._max_episode_steps:
            for i in self.agents:
                if i not in self._terminated_agents:
                    truncated[i] = True
            done = True
        else:
            for i, i_terminated in terminated.items():
                if i_terminated:
                    self._terminated_agents.add(i)

        return obs, rewards, terminated, truncated, done, info

    def reset(self, **kwargs):
        self._elapsed_steps = 0
        obs, info = self.env.reset(**kwargs)
        self._terminated_agents = set()
        return obs, info



"""Policies for the DrivingGen environments."""
from posggym.agents.grid_world.driving_gen.shortest_path import (
    DrivingGenShortestPathPolicy,
)
from posggym.agents.registration import PolicySpec

policy_specs = {}
for a, description in [
    (
        0.0,
        (
            "Follows shortest path to destination, doesn't go full speed, and stops if "
            "it observes another a car at any distance (aggressiveness=0.0)"
        ),
    ),
    (
        0.4,
        (
            "Follows shortest path to destination, doesn't go full speed, and stops if "
            "it observes a car near edge of observation range (aggressiveness=0.40)"
        ),
    ),
    (
        0.6,
        (
            "Follows shortest path to destination, goes up to full speed, and stops if "
            "it observes a car a medium distance away (aggressiveness=0.60)"
        ),
    ),
    (
        0.8,
        (
            "Follows shortest path to destination, goes up to full speed, and stops if "
            "it observes a car a very short distance away (aggressiveness=0.80)"
        ),
    ),
    (
        1.0,
        (
            "Follows shortest path to destination, goes up to full speed, and ignores "
            "other vehicles (aggressiveness=1.0)"
        ),
    ),
]:
    name = f"A{a*100:.0f}"
    spec = PolicySpec(
        policy_name=f"{name}Shortestpath",
        entry_point=DrivingGenShortestPathPolicy,
        version=0,
        env_id="DrivingGen-v1",
        env_args=None,
        valid_agent_ids=None,
        nondeterministic=False,
        kwargs={"aggressiveness": a},
        description=description,
    )
    policy_specs[spec.id] = spec



"""Functions and classes for registering and loading implemented agents.

Based on Farama Foundation Gymnasium,
https://github.com/Farama-Foundation/Gymnasium/blob/v0.27.0/gymnasium/envs/registration.py

"""
from __future__ import annotations

import copy
import difflib
import importlib
import re
from collections import defaultdict
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any, Dict, List, Protocol, Tuple

import posggym
from posggym import error, logger


if TYPE_CHECKING:
    import posggym.model as M
    from posggym.agents.policy import Policy


# [env_id/][env_args_id/](policy_id)-v(version)
# env_id is group 1, env_args_id is group 2. policy_id is group 2, version is group 3
POLICY_ID_RE: re.Pattern = re.compile(
    r"^((?:(?P<env_id>[\w:-]+)\/)?(?:(?P<env_args_id>[\S]+)\/)?)?"
    r"(?:(?P<pi_name>[\w:.-]+?))(?:-v(?P<version>\d+))?$"
)


class PolicyEntryPoint(Protocol):
    """Entry point function for instantiating a new policy instance."""

    def __call__(
        self, model: M.POSGModel, agent_id: str, policy_id: str, **kwargs
    ) -> Policy:
        ...


def load(name: str) -> PolicyEntryPoint:
    """Loads policy with name and returns a policy entry point.

    Arguments
    ---------
    name
        The policy name

    Returns
    -------
    entry_point
        Policy creation function.

    """
    mod_name, attr_name = name.split(":")
    mod = importlib.import_module(mod_name)
    fn = getattr(mod, attr_name)
    return fn


def parse_policy_id(policy_id: str) -> Tuple[str | None, str | None, str, int | None]:
    """Parse policy ID string format.

    env_id is group 1, env_args_id is group 2. policy_id is group 2, version is group 3

    [env_id/][env_args_id/](policy_id)-v(version)

    Arguments
    ---------
    policy_id
        The policy id to parse

    Returns
    -------
    env_id
        The environment ID
    env_args_id
        ID string of environment arguments
    pi_name
        The policy name
    version
        The policy version

    Raises
    ------
    Error
        If the policy id does not a valid environment regex

    """
    match = POLICY_ID_RE.fullmatch(policy_id)
    if not match:
        raise error.Error(
            f"Malformed policy ID: {policy_id}. Currently all IDs must be of the "
            "form [env_id/][env_args_id/](policy_name)-v(version) (env_id and "
            "env_args_id may be optional, depending on the policy)."
        )
    env_id, env_args_id, pi_name, version = match.group(
        "env_id", "env_args_id", "pi_name", "version"
    )
    if env_args_id is not None and env_id is None:
        raise error.Error(
            f"Malformed policy ID: {policy_id}. env_args_id is only valid if a valid "
            "if a valid env_id if included in the policy ID. Did not find valid env_id "
            f"and got env_args_id: {env_args_id}. Check env_id is correct."
        )

    if version is not None:
        version = int(version)

    return env_id, env_args_id, pi_name, version


def get_policy_id(
    env_id: str | None, env_args_id: str | None, policy_name: str, version: int | None
) -> str:
    """Get the full policy ID given a name and (optional) version and env-name.

    Inverse of :meth:`parse_policy_id`.

    Arguments
    ---------
    env_id
        The environment ID
    env_args_id
        ID string of environment arguments
    policy_name
        The policy name
    version
        The policy version

    Returns
    -------
    policy_id
        The policy id

    """
    if env_args_id is not None and env_id is None:
        raise error.Error(
            "Cannot create policy ID. Must include env_id if env_args_id is part of "
            "the policy ID."
        )

    full_name = policy_name
    if version is not None:
        full_name += f"-v{version}"
    if env_args_id is not None:
        full_name = env_args_id + "/" + full_name
    if env_id is not None:
        full_name = env_id + "/" + full_name
    return full_name


def get_env_args_id(env_args: Dict[str, Any]) -> str:
    """Get string representation of environment keyword arguments.

    Converts keyword dictionary {k1: v1, k2: v2, k3: v3} into a string:

        k1=v1-k2=v2-k3=v3

    Note we assume keywords are valid python variable names and so do not contain
    any hyphen '-' characters.

    Arguments
    ---------
    env_args
        environment keyword arguments

    Returns
    -------
    env_args_id
        string representation of the envrinment keyword arguments.

    """
    return ("-".join(f"{k}={v}" for k, v in env_args.items())).replace(" ", "")


@dataclass
class PolicySpec:
    """A specification for a particular agent policy.

    Used to register agent policies that can then be dynamically loaded using
    posggym_agents.make.

    Arguments
    ---------
    policy_name
        The name of the policy
    entry_point
        The Python entrypoint for initializing an instance of the agent policy.
        Must be a Callable with signature matching `PolicyEntryPoint` or a string
        defining where the entry point function can be imported
        (e.g. module.name:Class).
    version
        the policy version
    env_id
        Optional ID of the posggym environment that the policy is for.
    env_args
        Optional keywords arguments for the environment that the policy is for (if
        it is a environment specific policy). If None then assumes policy can be used
        for the environment with any arguments.
    env_args_id
        Optional string representation of the environment keyword arguments. If
        None then will generate an ID from the env_args.
    valid_agent_ids
        Optional AgentIDs in environment that policy is compatible with. If
        None then assumes policy can be used for any agent in the environment.
    nondeterministic
        Whether this policy is non-deterministic even after seeding.
    description
        Optional description of the policy.
    kwargs
        Additional kwargs, if any, to pass to the agent initializing


    Additional Attributes
    ---------------------
    id
        The unique policy identifier made from the env_id, env_args, policy_name, and
        version. Is of the form.
    env_args_id
        String representation of the env_args

    """

    policy_name: str
    entry_point: PolicyEntryPoint | str
    version: int | None = field(default=None)

    # Environment attributes
    env_id: str | None = field(default=None)
    env_args: Dict[str, Any] | None = field(default=None)

    # Policy attributes
    valid_agent_ids: List[str] | None = field(default=None)
    nondeterministic: bool = field(default=False)
    description: str | None = field(default=None)

    # Policy Arguments
    kwargs: Dict = field(default_factory=dict)

    # post-init attributes
    env_args_id: str | None = field(default=None)
    # the unique identifier for the policy spec
    id: str = field(init=False)

    def __post_init__(self):
        """Generate unique spec ID.

        Is called after spec is created.
        """
        if self.env_args is not None and self.env_args_id is None:
            self.env_args_id = get_env_args_id(self.env_args)

        # the unique ID for the policy spec
        self.id = get_policy_id(
            self.env_id, self.env_args_id, self.policy_name, self.version
        )
        # check id is valid
        parse_policy_id(self.id)

        if isinstance(self.valid_agent_ids, list) and len(self.valid_agent_ids) == 0:
            raise error.PolicyRegistrationError(
                f"Invalid PolicySpec for policy with id='{self.id}'. `valid_agent_ids` "
                "must be None or a non-empty list."
            )


def _check_env_id_exists(env_id: str | None, env_args_id: str | None):
    """Check if a env ID exists. If it doesn't, print a helpful error message."""
    if env_id is None:
        return

    env_ids = {spec_.env_id for spec_ in registry.values() if spec_.env_id is not None}
    if env_id not in env_ids:
        suggestion = (
            difflib.get_close_matches(env_id, env_ids, n=1)
            if len(env_ids) > 0
            else None
        )
        suggestion_msg = (
            f"Did you mean: `{suggestion[0]}`?"
            if suggestion
            else f"Have you installed the proper package for {env_id}?"
        )
        raise error.PolicyEnvIDNotFound(
            f"Environment ID {env_id} not found. {suggestion_msg}"
        )

    if env_args_id is None:
        return
    env_args_ids = {
        spec_.env_args_id
        for spec_ in registry.values()
        if (spec_.env_id == env_id and spec_.env_args_id is not None)
    }
    if env_args_id not in env_args_ids:
        suggestion = (
            difflib.get_close_matches(env_args_id, env_args_ids, n=1)
            if len(env_args_ids) > 0
            else None
        )
        suggestion_msg = f"Did you mean: `{suggestion[0]}`?" if suggestion else ""
        raise error.PolicyEnvArgsIDNotFound(
            f"Environment Arguments {env_args_id} for environment ID {env_id} not "
            f"found. {suggestion_msg}"
        )


def _check_name_exists(env_id: str | None, env_args_id: str | None, policy_name: str):
    """Check if policy exists for given env id. If not, print helpful error message."""
    # check if policy_name matches a generic policy
    names = {
        spec_.policy_name.lower(): spec_.policy_name
        for spec_ in registry.values()
        if spec_.env_id is None
    }
    if policy_name in names.values():
        return

    # check env specific policies
    _check_env_id_exists(env_id, env_args_id)
    names = {
        spec_.policy_name.lower(): spec_.policy_name
        for spec_ in registry.values()
        if (
            (spec_.env_id == env_id)
            and (env_args_id is None or spec_.env_args_id == env_args_id)
        )
    }
    if policy_name in names.values():
        return

    suggestion = difflib.get_close_matches(policy_name.lower(), names, n=1)
    env_id_msg = f" for env ID {env_id}" if env_id else ""
    suggestion_msg = f"Did you mean: `{names[suggestion[0]]}`?" if suggestion else ""

    raise error.PolicyNameNotFound(
        f"Policy {policy_name} doesn't exist{env_id_msg}. {suggestion_msg}"
    )


def _check_version_exists(
    env_id: str | None, env_args_id: str | None, policy_name: str, version: int | None
):
    """Check if policy version exists for env ID. Print helpful error message if not.

    This is a complete test whether an policy ID is valid, and will provide the best
    available hints.

    Arguments
    ---------
    env_id
        The environment ID
    env_args_ud
        The ID of the environment keyword arguments
    policy_name
        The policy name
    version
        The policy version

    Raises
    ------
    DeprecatedPolicy
        The policy doesn't exist but a default version does or the policy version is
        deprecated
    VersionNotFound
        The ``version`` used doesn't exist

    """
    if get_policy_id(env_id, env_args_id, policy_name, version) in registry:
        return

    _check_name_exists(env_id, env_args_id, policy_name)
    if version is None:
        return

    message = (
        f"Policy version `v{version}` for policy "
        f"`{get_policy_id(env_id, env_args_id, policy_name, None)}` doesn't exist."
    )

    policy_specs = [
        spec_
        for spec_ in registry.values()
        if (
            spec_.env_id == env_id
            and spec_.env_args_id == env_args_id
            and spec_.policy_name == policy_name
        )
    ]
    policy_specs = sorted(policy_specs, key=lambda spec_: int(spec_.version or -1))

    default_spec = [spec_ for spec_ in policy_specs if spec_.version is None]
    if default_spec:
        message += f" It provides the default version {default_spec[0].id}`."
        if len(policy_specs) == 1:
            raise error.DeprecatedPolicy(message)

    # Process possible versioned environments
    versioned_specs = [spec_ for spec_ in policy_specs if spec_.version is not None]

    latest_spec = max(
        versioned_specs, key=lambda spec: spec.version, default=None  # type: ignore
    )
    if latest_spec is None or latest_spec.version is None:
        return

    if version > latest_spec.version:
        version_list_msg = ", ".join(f"`v{spec_.version}`" for spec_ in policy_specs)
        message += f" It provides versioned policies: [ {version_list_msg} ]."
        raise error.PolicyVersionNotFound(message)

    if version < latest_spec.version:
        raise error.DeprecatedPolicy(
            f"Policy version v{version} for "
            f"`{get_policy_id(env_id, env_args_id, policy_name, None)}` "
            f"is deprecated. Please use `{latest_spec.id}` instead."
        )


def find_highest_version(
    env_id: str | None, env_args_id: str | None, policy_name: str
) -> int | None:
    """Finds the highest registered version of the policy in the registry."""
    version: list[int] = [
        spec_.version
        for spec_ in registry.values()
        if (
            spec_.env_id == env_id
            and spec_.env_args_id == env_args_id
            and spec_.policy_name == policy_name
            and spec_.version is not None
        )
    ]
    return max(version, default=None)


# Global registry of policies. Meant to be accessed through `register` and `make`
registry: dict[str, PolicySpec] = {}


def _check_spec_register(spec: PolicySpec):
    """Checks whether spec is valid to be registered.

    Helper function for `register`.
    """
    global registry
    latest_versioned_spec = max(
        (
            spec_
            for spec_ in registry.values()
            if (
                spec_.env_id == spec.env_id
                and spec_.env_args_id == spec.env_args_id
                and spec_.policy_name == spec.policy_name
                and spec_.version is not None
            )
        ),
        key=lambda spec_: int(spec_.version),  # type: ignore
        default=None,
    )

    unversioned_spec = next(
        (
            spec_
            for spec_ in registry.values()
            if (
                spec_.env_id == spec.env_id
                and spec_.env_args_id == spec.env_args_id
                and spec_.policy_name == spec.policy_name
                and spec_.version is None
            )
        ),
        None,
    )

    if unversioned_spec is not None and spec.version is not None:
        raise error.PolicyRegistrationError(
            "Can't register the versioned policy "
            f"`{spec.id}` when the unversioned policy "
            f"`{unversioned_spec.id}` of the same name already exists."
        )
    elif latest_versioned_spec is not None and spec.version is None:
        raise error.PolicyRegistrationError(
            "Can't register the unversioned policy "
            f"`{spec.id}` when the versioned policy "
            f"`{latest_versioned_spec.id}` of the same name "
            f"already exists. Note: the default behavior is "
            f"that `posggym_agents.make` with the unversioned policy "
            f"will return the latest versioned policy"
        )


def register(
    policy_name: str,
    entry_point: PolicyEntryPoint | str,
    version: int | None = None,
    env_id: str | None = None,
    env_args: Dict[str, Any] | None = None,
    valid_agent_ids: List[str] | None = None,
    nondeterministic: bool = False,
    description: str | None = None,
    **kwargs,
):
    """Register a policy with posggym.agents.

    The policy is registered in posggym so it can be used with
    :py:method:`posggym.agents.make`

    Arguments
    ---------
    policy_name
        The name of the policy
    entry_point
        The entry point for creating the policy
    env_id
        Optional ID of the posggym environment that the policy is for.
    version
        the policy version
    env_args
        Optional keywords arguments for the environment that the policy is for (if
        it is a environment specific policy). If None then assumes policy can be used
        for the environment with any arguments.
    valid_agent_ids
        Optional AgentIDs in environment that policy is compatible with. If
        None then assumes policy can be used for any agent in the environment.
    nondeterministic
        Whether this policy is non-deterministic even after seeding.
    description
        Optional description of the policy.
    kwargs
        Additional kwargs, if any, to pass to the agent initializing

    """
    global registry
    new_spec = PolicySpec(
        policy_name=policy_name,
        entry_point=entry_point,
        version=version,
        env_id=env_id,
        env_args=env_args,
        valid_agent_ids=valid_agent_ids,
        nondeterministic=nondeterministic,
        description=description,
        **kwargs,
    )
    register_spec(new_spec)


def register_spec(spec: PolicySpec):
    """Register a policy spec with posggym-agents.

    Arguments
    ---------
    spec
        The policy spec

    """
    global registry
    _check_spec_register(spec)
    if spec.id in registry:
        logger.warn(f"Overriding policy {spec.id} already in registry.")
    registry[spec.id] = spec


def make(id: str | PolicySpec, model: M.POSGModel, agent_id: str, **kwargs) -> Policy:
    """Create an policy according to the given ID.

    To find all available policies use `posggym_agents.agents.registry.keys()` for
    all valid ids.

    Arguments
    ---------
    id
      Unique identifier of the policy or a policy spec.
    model
      The model for the environment the policy will be interacting with.
    agent_id
      The ID of the agent the policy will be used for.
    kwargs
      Additional arguments to pass to the policy constructor.

    Returns
    -------
    policy
      An instance of the policy.

    Raises
    ------
    Error
      If the ``id`` doesn't exist then an error is raised

    """
    if isinstance(id, PolicySpec):
        spec_: PolicySpec = id
    else:
        env_id, env_args_id, policy_name, version = parse_policy_id(id)

        if id not in registry:
            generic_names = {
                spec.policy_name for spec in registry.values() if spec.env_id is None
            }
            if policy_name in generic_names:
                env_id = None
                env_args_id = None

        policy_id = get_policy_id(env_id, env_args_id, policy_name, version)
        spec_ = registry.get(policy_id)  # type: ignore

        latest_version = find_highest_version(env_id, env_args_id, policy_name)
        if (
            version is not None
            and latest_version is not None
            and latest_version > version
        ):
            logger.warn(
                f"The policy {id} is out of date. You should consider "
                f"upgrading to version `v{latest_version}`."
            )

        if version is None and latest_version is not None:
            version = latest_version
            new_policy_id = get_policy_id(env_id, env_args_id, policy_name, version)
            spec_ = registry.get(new_policy_id)  # type: ignore
            logger.warn(
                f"Using the latest versioned policy `{new_policy_id}` "
                f"instead of the unversioned policy `{id}`."
            )

        if spec_ is None:
            _check_version_exists(env_id, env_args_id, policy_name, version)
            raise error.Error(f"No registered policy with id: {id}")

    if spec_.valid_agent_ids and agent_id not in spec_.valid_agent_ids:
        raise error.Error(
            f"Attempted to initialize policy with ID={spec_.id} with invalid "
            f"agent ID '{agent_id}'. Valid agent IDs for this policy are: "
            f"{spec_.valid_agent_ids}."
        )

    _kwargs = spec_.kwargs.copy()
    _kwargs.update(kwargs)

    if spec_.entry_point is None:
        raise error.Error(f"{spec_.id} registered but entry_point is not specified")
    elif callable(spec_.entry_point):
        policy_creator = spec_.entry_point
    else:
        policy_creator = load(spec_.entry_point)

    try:
        policy = policy_creator(model, agent_id, spec_.id, **_kwargs)
    except TypeError as e:
        raise e

    # Copies the policy creation specification and kwargs to add to the
    # policy's specification details
    spec_ = copy.deepcopy(spec_)
    spec_.kwargs = _kwargs
    policy.spec = spec_

    return policy


def spec(id: str) -> PolicySpec:
    """Retrieve the spec for the given policy from the global registry.

    Arguments
    ---------
    id
        the policy id.

    Returns
    -------
    spec
        the policy spec from the global registry.

    Raises
    ------
    Error
        if policy with given ``id`` doesn't exist in global registry.

    """
    spec_ = registry.get(id)
    if spec_ is None:
        env_id, env_args_id, policy_name, version = parse_policy_id(id)
        _check_version_exists(env_id, env_args_id, policy_name, version)

        # No error raised so policy_name-version may be generic
        id = get_policy_id(None, None, policy_name, version)
        _check_version_exists(None, None, policy_name, version)
        spec_ = registry.get(id)

    if spec_ is None:
        raise error.Error(f"No registered policy with id: {id}")
    else:
        assert isinstance(spec_, PolicySpec)
        return spec_


def pprint_registry(
    _registry: dict = registry,
    num_cols: int = 3,
    include_env_ids: List[str] | None = None,
    exclude_env_ids: List[str] | None = None,
    disable_print: bool = False,
) -> str | None:
    """Pretty print the policies in the registry.

    Arguments
    ---------
    _registry
        Policy registry to be printed.
    num_cols
        Number of columns to arrange policies in, for display.
    include_env_ids
        Print only policies for environments with these IDs. If None then all
        environments are included.
    exclude_env_ids
        Exclude any policies for environments with thee IDs from being printed.
    disable_print
        Whether to return a string of all the policy IDs instead of printing it to
        console.

    Returns
    -------
    return_str
        formatted str representation of registry, if ``disable_print=True``, otherwise
        returns ``None``.

    """
    # Defaultdict to store policy names according to env_id.
    env_policies = defaultdict(lambda: defaultdict(lambda: []))
    max_justify = 0
    for spec in _registry.values():
        env_id = "Generic" if spec.env_id is None else spec.env_id
        env_policies[env_id][spec.env_args_id].append(
            f"{spec.policy_name}-v{spec.version}"
        )
        max_justify = max(max_justify, len(f"{spec.policy_name}-v{spec.version}"))

    # Iterate through each environment and print policies alphabetically.
    return_str = ""
    for env_id in env_policies:
        if exclude_env_ids is not None and env_id in exclude_env_ids:
            continue
        if include_env_ids is not None and env_id not in include_env_ids:
            continue

        return_str += f"{'=' * 5} {env_id} {'=' * 5}\n"
        for env_args_id, policies in env_policies[env_id].items():
            if env_args_id is not None:
                return_str += f"{'-' * 5} {env_id}/{env_args_id} {'-' * 5}\n"
            # Reference: https://stackoverflow.com/a/33464001
            for count, item in enumerate(sorted(policies), 1):
                return_str += (
                    item.ljust(max_justify) + " "
                )  # Print column with justification.
                # Once all rows printed, switch to new column.
                if count % num_cols == 0 or count == len(policies):
                    return_str = return_str.rstrip(" ") + "\n"
            return_str += "\n"
        return_str += "\n"

    if disable_print:
        return return_str

    print(return_str, end="")
    return None


def get_all_env_policies(
    env_id: str,
    env_args: Dict[str, Any] | str | None = None,
    _registry: Dict = registry,
    include_generic_policies: bool = True,
) -> List[PolicySpec]:
    """Get all PolicySpecs that are associated with a given environment ID.

    Arguments
    ---------
    env_id
        The ID of the environment
    env_args
        Optional environment arguments or ID string of environment arguments. If
        None, will return all policies for given environment.
    _registry
        The policy registry
    include_generic_policies
        whether to also return policies that are valid for all environments (e.g. the
        random-v0 policy)

    Returns
    -------
    policy_specs
        list of specs for policies associated with given environment.

    """
    return [
        spec
        for spec in _registry.values()
        if (
            (include_generic_policies and spec.env_id is None)
            or (
                spec.env_id == env_id
                and (
                    env_args is None
                    or spec.env_args_id is None
                    or (isinstance(env_args, str) and spec.env_args_id == env_args)
                    or (isinstance(env_args, dict) and spec.env_args == env_args)
                )
            )
        )
    ]


def get_env_agent_policies(
    env_id: str,
    env_args: Dict[str, Any] | None = None,
    _registry: Dict = registry,
    include_generic_policies: bool = True,
) -> Dict[str, List[PolicySpec]]:
    """Get each agent's policy specs associated with given environment.

    Arguments
    ---------
    env_id
        The ID of the environment
    env_args
        Optional environment arguments. If None, will return all policies for
        given environment.
    _registry
        The policy registry
    include_generic_policies
        whether to also return policies that are valid for all environments (e.g. the
        random-v0 policy) and environment args

    Returns
    -------
    policy_specs
        list of specs for policies associated with given environment.

    """
    env = posggym.make(env_id) if env_args is None else posggym.make(env_id, **env_args)

    policies: Dict[str, List[PolicySpec]] = {i: [] for i in env.possible_agents}
    for spec in get_all_env_policies(
        env_id,
        env_args,
        _registry=_registry,
        include_generic_policies=include_generic_policies,
    ):
        for i in env.possible_agents:
            if spec.valid_agent_ids is None or i in spec.valid_agent_ids:
                policies[i].append(spec)
    return policies


def get_all_envs(
    _registry: Dict = registry,
) -> Dict[str, Dict[str | None, Dict[str, Any] | None]]:
    """Get all the environments that have at least one registered policy.

    Arguments
    ---------
    _registry
        The policy registry

    Returns
    -------
    envs
        a dictionary with env IDs as keys as list of (env_args, env_args_id) tuples as
        the values.

    """
    envs: Dict[str, Dict[str | None, Dict[str, Any] | None]] = {}
    for spec in _registry.values():
        if spec.env_id is not None:
            envs.setdefault(spec.env_id, {})
            envs[spec.env_id][spec.env_args_id] = spec.env_args
    return envs



"""Shortest path policy for PursuitEvasion env."""
from __future__ import annotations

from queue import PriorityQueue
from typing import TYPE_CHECKING, Dict, Set, Tuple, cast

import posggym.envs.grid_world.pursuit_evasion as pe
from posggym.agents.policy import Policy, PolicyID, PolicyState
from posggym.agents.utils import action_distributions
from posggym.envs.grid_world.core import Coord, Direction
from posggym.utils import seeding

if TYPE_CHECKING:
    from posggym.model import POSGModel
    from posggym.utils.history import AgentHistory


# Current coord, facing direction
Pos = Tuple[Coord, Direction]


class PEShortestPathPolicy(Policy[pe.PEAction, pe.PEObs]):
    """Shortest Path Policy for the Pursuit Evasion environment.

    This policy sets the preferred action as the one which is on the shortest
    path to a target location evaders goal.

    For the evader the target location is the evader's goal.

    For the pursuer the target location is the evader's start location. If the pursuer
    reaches the evader's start location without the episode ending, the target location
    changes to the next closest unexplored evader start or goal location.

    """

    def __init__(self, model: POSGModel, agent_id: str, policy_id: PolicyID):
        super().__init__(model, agent_id, policy_id)
        self.model = cast(pe.PursuitEvasionModel, model)
        self._grid = self.model.grid
        self._is_evader = int(self.agent_id) == self.model.EVADER_IDX
        self._action_space = list(range(self.model.action_spaces[self.agent_id].n))

        self.shortest_paths = {}
        for dest_coord in set(
            self._grid.evader_start_coords + self._grid.all_goal_coords
        ):
            self.shortest_paths[dest_coord] = self.get_all_shortest_paths(dest_coord)

        self._rng, _ = seeding.std_random()

    def reset(self, *, seed: int | None = None):
        super().reset(seed=seed)
        if seed is not None:
            self._rng, _ = seeding.std_random(seed=seed)

    def get_initial_state(self) -> PolicyState:
        state = super().get_initial_state()
        state.update(
            {
                "update_num": 0,
                "facing_dir": pe.INITIAL_DIR,
                "coord": (0, 0),
                "prev_coord": (0, 0),
                "target_coord": (0, 0),
                "visited_target_coords": None,
            }
        )
        return state

    def get_next_state(
        self,
        action: pe.PEAction | None,
        obs: pe.PEObs,
        state: PolicyState,
    ) -> PolicyState:
        if state["update_num"] == 0:
            # I.e. first observation with no action performed yet
            next_update_num = 1
            next_facing_dir = pe.INITIAL_DIR
            next_coord = obs[1] if self._is_evader else obs[2]
            next_prev_coord = (0, 0)
            target_coord = obs[3] if self._is_evader else obs[1]
            visited_target_coords = None
        else:
            next_update_num = state["update_num"] + 1
            next_facing_dir = pe.ACTION_TO_DIR[action][state["facing_dir"]]
            next_coord = self._grid.get_next_coord(
                state["coord"], next_facing_dir, False
            )
            next_prev_coord = state["coord"]
            target_coord = state["target_coord"]

            if not self._is_evader and next_coord == target_coord:
                if state["visited_target_coords"] is None:
                    visited_target_coords = []
                else:
                    visited_target_coords = list(state["visited_target_coords"])

                visited_target_coords.append(target_coord)
                unvisited_target_coords = set(
                    self._grid.evader_start_coords + self._grid.all_goal_coords
                ) - set(visited_target_coords)

                if unvisited_target_coords:
                    target_coord = min(
                        unvisited_target_coords,
                        key=lambda c: self.get_dest_shortest_path_dist(
                            c, (next_coord, next_facing_dir)
                        ),
                    )
                else:
                    # All evader start and goal locations have been visited so just keep
                    # target as current target (this is unlikely to happen within
                    # episode step limits)
                    target_coord = next_coord
                visited_target_coords = tuple(visited_target_coords)
            else:
                visited_target_coords = state["visited_target_coords"]

        next_state = {
            "update_num": next_update_num,
            "facing_dir": next_facing_dir,
            "coord": next_coord,
            "prev_coord": next_prev_coord,
            "target_coord": target_coord,
            "visited_target_coords": visited_target_coords,
        }
        return next_state

    def get_state_from_history(self, history: AgentHistory) -> PolicyState:
        last_action, obs = history.get_last_step()
        assert obs is not None
        facing_dir = pe.INITIAL_DIR
        coord = obs[1] if self._is_evader else obs[2]
        prev_coord = coord
        target_coord = obs[3] if self._is_evader else obs[1]
        for a, _ in history:
            if a is None:
                continue
            facing_dir = pe.ACTION_TO_DIR[a][facing_dir]
            prev_coord = coord
            coord = self._grid.get_next_coord(coord, facing_dir, False)
        return {
            "action": last_action,
            "update_num": len(history),
            "facing_dir": facing_dir,
            "coord": coord,
            "prev_coord": prev_coord,
            "target_coord": target_coord,
        }

    def sample_action(self, state: PolicyState) -> pe.PEAction:
        return self.get_pi(state).sample()

    def get_pi(self, state: PolicyState) -> action_distributions.ActionDistribution:
        dists = []
        coord, target_coord = state["coord"], state["target_coord"]
        for a in self._action_space:
            a_dir = pe.ACTION_TO_DIR[a][state["facing_dir"]]
            next_coord = self._grid.get_next_coord(coord, a_dir, False)
            d = self.get_dest_shortest_path_dist(target_coord, (next_coord, a_dir))
            dists.append(d)

        min_dist = min(dists)
        num_min = dists.count(min_dist)
        pi = {}
        for a in self._action_space:
            if dists[a] == min_dist:
                pi[a] = 1 / num_min
            else:
                pi[a] = 0

        return action_distributions.DiscreteActionDistribution(pi, self._rng)

    def get_value(self, state: PolicyState) -> float:
        raise NotImplementedError(
            f"`get_value()` no implemented by {self.__class__.__name__} policy"
        )

    def get_dest_shortest_path_dist(self, dest_coord: Coord, pos: Pos) -> int:
        """Get shortest path distance to destination coord from given position."""
        if dest_coord not in self.shortest_paths:
            self.shortest_paths[dest_coord] = self.get_all_shortest_paths(dest_coord)

        min_dist = float("inf")
        for dists in self.shortest_paths[dest_coord].values():
            min_dist = min(min_dist, dists.get(pos, float("inf")))
        return min_dist

    def get_all_shortest_paths(self, origin: Coord) -> Dict[Pos, Dict[Pos, int]]:
        """Get shortest paths from given origin to all other positions.

        Note, this is a search over agent configurations (coord, facing_dir), rather
        than just agent coordinate. This yields a shortest path distance in terms of
        number of actions, rather than number of cells.
        """
        src_dists = {}
        for facing_dir in Direction:
            pos = (origin, facing_dir)
            src_dists[pos] = self.dijkstra(pos)
        return src_dists

    def dijkstra(self, origin: Pos) -> Dict[Pos, int]:
        """Get shortest path distance to origin from all other positions."""
        dist = {origin: 0}
        pq = PriorityQueue()  # type: ignore
        pq.put((dist[origin], origin))

        visited = {origin}

        while not pq.empty():
            _, pos = pq.get()

            for adj_pos in self.get_prev_positions(pos):
                if dist[pos] + 1 < dist.get(adj_pos, float("inf")):
                    dist[adj_pos] = dist[pos] + 1

                    if adj_pos not in visited:
                        pq.put((dist[adj_pos], adj_pos))
                        visited.add(adj_pos)
        return dist

    def get_prev_positions(self, pos: Pos) -> Set[Pos]:
        """Get all positions reachable from given position."""
        coord, facing_dir = pos
        prev_positions = set()
        for prev_a in [pe.FORWARD, pe.BACKWARD, pe.LEFT, pe.RIGHT]:
            if prev_a == pe.FORWARD:
                prev_dir = facing_dir
                move_dir = Direction((facing_dir + 2) % len(Direction))
            elif prev_a == pe.BACKWARD:
                prev_dir = facing_dir
                move_dir = facing_dir
            elif prev_a == pe.LEFT:
                prev_dir = Direction((facing_dir + 1) % len(Direction))
                move_dir = Direction((facing_dir + 2) % len(Direction))
            else:
                # RIGHT
                prev_dir = Direction((facing_dir - 1) % len(Direction))
                move_dir = Direction((facing_dir + 2) % len(Direction))

            next_coord = self._grid.get_next_coord(
                coord,
                Direction((move_dir + 2) % len(Direction)),
                ignore_blocks=False,
            )
            if next_coord == coord:
                # if next coord is blocked then may have moved to current
                # coord from 0 or 1 cells away
                prev_positions.add((coord, prev_dir))

            prev_coord = prev_coord = self._grid.get_next_coord(
                coord, move_dir, ignore_blocks=False
            )
            prev_positions.add((prev_coord, prev_dir))

        return prev_positions



"""Policies for the PredatorPrey-v0 environment."""
import os.path as osp

from posggym.agents.grid_world.predator_prey import heuristic
from posggym.agents.registration import PolicySpec
from posggym.agents.torch_policy import PPOPolicy
from posggym.agents.utils import processors
from posggym.config import AGENT_MODEL_DIR

agent_model_dir = osp.join(AGENT_MODEL_DIR, "grid_world", "predator_prey")
policy_specs = {}

for policy_class in [
    heuristic.PPHeuristic1,
    heuristic.PPHeuristic2,
    heuristic.PPHeuristic3,
]:
    num = policy_class.__name__.replace("PPHeuristic", "")
    spec = PolicySpec(
        policy_name=f"H{num}",
        entry_point=policy_class,
        env_id="PredatorPrey-v0",
        env_args=None,
        version=0,
        valid_agent_ids=None,
        nondeterministic=False,
        description=policy_class.__doc__.replace("\n", " "),
    )
    policy_specs[spec.id] = spec

# PredatorPrey-10x10-P2-p3-s2-coop-v0
for policy_file_name in [
    "RL1.pkl",
    "RL2.pkl",
    "RL3.pkl",
    "RL4.pkl",
    "RL5.pkl",
    "RL6.pkl",
    "RL7.pkl",
    "RL8.pkl",
]:
    spec = PPOPolicy.get_spec_from_path(
        env_id="PredatorPrey-v0",
        env_args={
            "grid": "10x10",
            "num_predators": 2,
            "num_prey": 3,
            "cooperative": True,
            "prey_strength": 2,
            "obs_dim": 2,
        },
        env_args_id="grid=10x10-num_predators=2-num_prey=3-cooperative=True",
        policy_file_path=osp.join(
            agent_model_dir,
            "grid=10x10-num_predators=2-num_prey=3-cooperative=True",
            policy_file_name,
        ),
        version=0,
        valid_agent_ids=None,
        # policy is deterministic given random seed
        nondeterministic=False,
        # actions sampled, rather always taking most probable action
        deterministic=False,
        obs_processor_cls=processors.FlattenProcessor,
        obs_processor_config=None,
        action_processor_cls=None,
        action_processor_config=None,
        description="Deep RL policy trained using PPO and self-play.",
    )
    policy_specs[spec.id] = spec



"""Policies for the Driving environments."""
import os.path as osp

from posggym.agents.grid_world.driving.shortest_path import DrivingShortestPathPolicy
from posggym.agents.registration import PolicySpec
from posggym.agents.torch_policy import PPOPolicy
from posggym.agents.utils import processors
from posggym.config import AGENT_MODEL_DIR


agent_model_dir = osp.join(AGENT_MODEL_DIR, "grid_world", "driving")
policy_specs = {}


for a, description in [
    (
        0.0,
        (
            "Follows shortest path to destination, doesn't go full speed, and stops if "
            "it observes another a car at any distance (aggressiveness=0.0)"
        ),
    ),
    (
        0.4,
        (
            "Follows shortest path to destination, doesn't go full speed, and stops if "
            "it observes a car near edge of observation range (aggressiveness=0.40)"
        ),
    ),
    (
        0.6,
        (
            "Follows shortest path to destination, goes up to full speed, and stops if "
            "it observes a car a medium distance away (aggressiveness=0.60)"
        ),
    ),
    (
        0.8,
        (
            "Follows shortest path to destination, goes up to full speed, and stops if "
            "it observes a car a very short distance away (aggressiveness=0.80)"
        ),
    ),
    (
        1.0,
        (
            "Follows shortest path to destination, goes up to full speed, and ignores "
            "other vehicles (aggressiveness=1.0)"
        ),
    ),
]:
    name = f"A{a*100:.0f}"
    spec = PolicySpec(
        policy_name=f"{name}Shortestpath",
        entry_point=DrivingShortestPathPolicy,
        version=0,
        env_id="Driving-v1",
        env_args=None,
        valid_agent_ids=None,
        nondeterministic=False,
        kwargs={"aggressiveness": a},
        description=description,
    )
    policy_specs[spec.id] = spec


for policy_file_name in [
    "RL1.pkl",
    "RL2.pkl",
    "RL3.pkl",
    "RL4.pkl",
    "RL5.pkl",
]:
    spec = PPOPolicy.get_spec_from_path(
        env_id="Driving-v1",
        env_args={
            "grid": "14x14RoundAbout",
            "num_agents": 2,
            "obs_dim": (3, 1, 1),
        },
        env_args_id="grid=14x14RoundAbout-num_agents=2",
        policy_file_path=osp.join(
            agent_model_dir,
            "grid=14x14RoundAbout-num_agents=2",
            policy_file_name,
        ),
        version=0,
        valid_agent_ids=None,
        # policy is deterministic given random seed
        nondeterministic=False,
        # actions sampled, rather always taking most probable action
        deterministic=False,
        obs_processor_cls=processors.FlattenProcessor,
        obs_processor_config=None,
        action_processor_cls=None,
        action_processor_config=None,
        description="Deep RL policy trained using PPO and self-play.",
    )
    policy_specs[spec.id] = spec



"""Checks that the core posggym model API is implemented as expected."""
from typing import Dict, List

from gymnasium import spaces

import posggym.model as M
from posggym.utils import seeding


class ExampleModel(M.POSGModel[int, int, int]):
    """Example discrete testing model."""

    def __init__(self):
        self.possible_agents = (0, 1)
        self.action_spaces = {i: spaces.Discrete(2) for i in self.possible_agents}
        self.observation_spaces = {i: spaces.Discrete(2) for i in self.possible_agents}
        self.is_symmetric = True

    @property
    def rng(self) -> seeding.RNG:
        if self._rng is None:
            self._rng, seed = seeding.std_random()
        return self._rng

    def get_agents(self, state: int) -> List[str]:
        return list(self.possible_agents)

    def sample_initial_state(self) -> int:
        return 0

    def sample_initial_obs(self, state: int) -> Dict[str, int]:
        return {i: 0 for i in self.possible_agents}

    def step(self, state: int, actions: Dict[str, int]) -> M.JointTimestep[int, int]:
        return M.JointTimestep(
            0,
            {i: 0 for i in self.possible_agents},
            {i: 0 for i in self.possible_agents},
            {i: False for i in self.possible_agents},
            {i: False for i in self.possible_agents},
            False,
            {i: {} for i in self.possible_agents},
        )


def test_posggym_model():
    """Tests general posggym model API."""
    model = ExampleModel()

    assert model.state_space is None
    assert model.reward_ranges == {
        i: (-float("inf"), float("inf")) for i in model.possible_agents
    }
    assert model.spec is None
    assert model._rng is None  # pylint: disable=protected-access



"""Checks that the core posggym environment API is implemented as expected.

Adapted from:
https://github.com/Farama-Foundation/Gymnasium/blob/v0.27.0/tests/test_core.py

"""
from typing import Any, Dict, Optional, Tuple

import numpy as np
import posggym.model as M
from gymnasium.spaces import Box
from posggym import (
    ActionWrapper,
    DefaultEnv,
    Env,
    ObservationWrapper,
    RewardWrapper,
    Wrapper,
)
from posggym.core import WrapperActType, WrapperObsType

from tests.test_model import ExampleModel


class ExampleEnv(DefaultEnv[int, int, int]):
    """Example testing environment."""

    def __init__(self):
        super().__init__(ExampleModel())


def test_posggym_env():
    """Tests general posggym environment API."""
    env = ExampleEnv()

    assert env.metadata == {"render_modes": []}
    assert env.render_mode is None
    assert env.spec is None


class ExampleWrapper(Wrapper):
    """An example testing wrapper."""

    def __init__(self, env: Env[M.StateType, M.ObsType, M.ActType]):
        """Constructor that sets the reward."""
        super().__init__(env)
        self.new_reward = 3

    def reset(
        self, *, seed: Optional[int] = None, options: Optional[Dict[str, Any]] = None
    ) -> Tuple[Dict[str, WrapperObsType], Dict[str, Dict]]:
        return super().reset(seed=seed, options=options)

    def step(
        self, actions: Dict[str, WrapperActType]
    ) -> Tuple[
        Dict[str, WrapperObsType],
        Dict[str, float],
        Dict[str, bool],
        Dict[str, bool],
        bool,
        Dict[str, Dict],
    ]:
        obs, reward, term, trunc, done, info = self.env.step(actions)  # type: ignore
        reward = {i: self.new_reward for i in reward}
        return obs, reward, term, trunc, done, info


def test_posggym_wrapper():
    """Tests the posggym wrapper API works as expected."""
    env = ExampleEnv()
    wrapper_env = ExampleWrapper(env)

    assert env.metadata == wrapper_env.metadata
    wrapper_env.metadata = {"render_modes": ["rgb_array"]}
    assert env.metadata != wrapper_env.metadata

    assert env.render_mode == wrapper_env.render_mode

    assert env.reward_ranges == wrapper_env.reward_ranges
    wrapper_env.reward_ranges = {i: (-1.0, 1.0) for i in wrapper_env.possible_agents}
    assert env.reward_ranges != wrapper_env.reward_ranges

    assert env.spec == wrapper_env.spec

    assert env.observation_spaces == wrapper_env.observation_spaces
    assert env.action_spaces == wrapper_env.action_spaces
    wrapper_env.observation_spaces = {i: Box(1, 2) for i in wrapper_env.possible_agents}
    wrapper_env.action_spaces = {i: Box(1, 2) for i in wrapper_env.possible_agents}
    assert env.observation_spaces != wrapper_env.observation_spaces
    assert env.action_spaces != wrapper_env.action_spaces

    assert wrapper_env.model is env.model


class ExampleRewardWrapper(RewardWrapper):
    """Example reward wrapper for testing."""

    def rewards(self, rewards: Dict[str, float]) -> Dict[str, float]:
        return {i: 1 for i in rewards}


class ExampleObservationWrapper(ObservationWrapper):
    """Example observation wrapper for testing."""

    def observations(self, obs: Dict[str, M.ObsType]) -> Dict[str, WrapperObsType]:
        return {i: np.array([1]) for i in obs}  # type: ignore


class ExampleActionWrapper(ActionWrapper):
    """Example action wrapper for testing."""

    def actions(self, actions: Dict[str, M.ActType]) -> Dict[str, WrapperActType]:
        return {i: np.array([1]) for i in actions}  # type: ignore


class ActionWrapperTestEnv(DefaultEnv[int, int, int]):
    """Example environment for tesint action wrapper.

    Step returns the action as an observation.
    """

    def __init__(self):
        super().__init__(ExampleModel())

    def step(
        self, actions: Dict[str, int]
    ) -> Tuple[
        Dict[str, int],
        Dict[str, float],
        Dict[str, bool],
        Dict[str, bool],
        bool,
        Dict[str, Dict],
    ]:
        step = self.model.step(self._state, actions)
        self._step_num += 1
        self._state = step.state
        self._last_obs = step.observations
        self._last_actions = actions
        self._last_rewards = step.rewards
        return (
            actions,
            step.rewards,
            step.terminations,
            step.truncations,
            step.all_done,
            step.infos,
        )


def test_wrapper_types():
    """Tests the observation, action and reward wrapper examples."""
    env = ExampleEnv()

    reward_env = ExampleRewardWrapper(env)
    reward_env.reset()
    _, rewards, _, _, _, _ = reward_env.step(0)
    assert all(r_i == 1 for r_i in rewards.values())

    observation_env = ExampleObservationWrapper(env)
    obs, _ = observation_env.reset()
    assert all(o_i == np.array([1]) for o_i in obs.values())
    obs, _, _, _, _, _ = observation_env.step({i: 0 for i in env.possible_agents})
    assert all(o_i == np.array([1]) for o_i in obs.values())

    env = ActionWrapperTestEnv()
    action_env = ExampleActionWrapper(env)
    obs, _, _, _, _, _ = action_env.step({i: 0 for i in env.possible_agents})
    assert all(o_i == np.array([1]) for o_i in obs.values())



"""General grid world problem utility functions and classes."""
import enum
import itertools
import random
from queue import PriorityQueue, Queue
from typing import Dict, Iterable, List, Optional, Set, Tuple

# (x, y) coord = (col, row) coord
Coord = Tuple[int, int]


class Direction(enum.IntEnum):
    """A direction in a grid."""

    NORTH = 0
    EAST = 1
    SOUTH = 2
    WEST = 3


DIRECTION_ASCII_REPR = ["^", ">", "v", "<"]


class Grid:
    """A base grid class, with general utility functions and attributes."""

    def __init__(
        self,
        grid_width: int,
        grid_height: int,
        block_coords: Optional[Set[Coord]] = None,
    ):
        self.width = grid_width
        self.height = grid_height

        if block_coords is None:
            block_coords = set()
        self.block_coords = block_coords

    @property
    def all_coords(self) -> List[Coord]:
        """The list of all locations on grid, including blocks."""
        return list(itertools.product(range(self.width), range(self.width)))

    @property
    def n_coords(self) -> int:
        """The total number of coordinates on grid, including blocks."""
        return self.height * self.width

    @property
    def unblocked_coords(self) -> List[Coord]:
        """The list of all coordinates on the grid excluding blocks."""
        return [coord for coord in self.all_coords if coord not in self.block_coords]

    @staticmethod
    def manhattan_dist(coord1: Coord, coord2: Coord) -> int:
        """Get manhattan distance between two coordinates on the grid."""
        return abs(coord1[0] - coord2[0]) + abs(coord1[1] - coord2[1])

    def coord_in_bounds(self, coord: Coord) -> bool:
        """Return whether a coordinate is inside the grid or not."""
        return 0 <= coord[0] < self.width and 0 <= coord[1] < self.height

    def get_neighbours(
        self,
        coord: Coord,
        ignore_blocks: bool = False,
        include_out_of_bounds: bool = False,
    ) -> List[Coord]:
        """Get set of adjacent non-blocked coordinates."""
        neighbours = []
        if coord[1] > 0 or include_out_of_bounds:
            neighbours.append((coord[0], coord[1] - 1))  # N
        if coord[0] < self.width - 1 or include_out_of_bounds:
            neighbours.append((coord[0] + 1, coord[1]))  # E
        if coord[1] < self.height - 1 or include_out_of_bounds:
            neighbours.append((coord[0], coord[1] + 1))  # S
        if coord[0] > 0 or include_out_of_bounds:
            neighbours.append((coord[0] - 1, coord[1]))  # W

        if ignore_blocks:
            return neighbours

        for i in range(len(neighbours), 0, -1):
            if neighbours[i - 1] in self.block_coords:
                neighbours.pop(i - 1)

        return neighbours

    def get_next_coord(
        self, coord: Coord, move_dir: Direction, ignore_blocks: bool = False
    ) -> Coord:
        """Get next coordinate given loc and movement direction.

        If new coordinate is outside of the grid boundary, or (ignore_blocks is
        False and new coordinate is a block) then returns the original
        coordinate.
        """
        new_coord_list = list(coord)
        if move_dir == Direction.NORTH:
            new_coord_list[1] = max(0, coord[1] - 1)
        elif move_dir == Direction.EAST:
            new_coord_list[0] = min(self.width - 1, coord[0] + 1)
        elif move_dir == Direction.SOUTH:
            new_coord_list[1] = min(self.height - 1, coord[1] + 1)
        elif move_dir == Direction.WEST:
            new_coord_list[0] = max(0, coord[0] - 1)
        new_coord = (new_coord_list[0], new_coord_list[1])

        if not ignore_blocks and new_coord in self.block_coords:
            return coord
        return new_coord

    def get_coords_within_dist(
        self, origin: Coord, dist: int, ignore_blocks: bool, include_origin: bool
    ) -> Set[Coord]:
        """Get set of coords within given distance from origin."""
        if dist == 0:
            return {origin} if include_origin else set()

        adj_coords = self.get_neighbours(origin, ignore_blocks)
        in_dist_coords = set(adj_coords)

        if include_origin and (ignore_blocks or origin not in self.block_coords):
            in_dist_coords.add(origin)

        if dist == 1:
            return in_dist_coords

        for coord in adj_coords:
            in_dist_coords.update(
                self.get_coords_within_dist(coord, dist - 1, ignore_blocks, False)
            )

        if not include_origin and origin in in_dist_coords:
            # must remove since it will get added again during recursive call
            in_dist_coords.remove(origin)

        return in_dist_coords

    def get_coords_at_dist(
        self, origin: Coord, dist: int, ignore_blocks: bool
    ) -> Set[Coord]:
        """Get set of coords at given distance from origin."""
        if dist == 0:
            return {origin}

        in_dist_coords = self.get_coords_within_dist(origin, dist, ignore_blocks, False)

        at_dist_coords: Set[Coord] = set()
        for coord in in_dist_coords:
            if self.manhattan_dist(origin, coord) == dist:
                at_dist_coords.add(coord)

        return at_dist_coords

    def get_min_dist_coords(
        self, origin: Coord, coords: Iterable[Coord]
    ) -> List[Coord]:
        """Get list of coord in coords closest to origin."""
        dists = self.get_coords_by_distance(origin, coords)
        if len(dists) == 0:
            return []
        return dists[min(dists)]

    def get_coords_by_distance(
        self, origin: Coord, coords: Iterable[Coord]
    ) -> Dict[int, List[Coord]]:
        """Get mapping from distance to coords at that distance from origin."""
        dists: Dict[int, List[Coord]] = {}
        for coord in coords:
            d = self.manhattan_dist(origin, coord)
            if d not in dists:
                dists[d] = []
            dists[d].append(coord)
        return dists

    def get_all_shortest_paths(
        self, origins: Iterable[Coord]
    ) -> Dict[Coord, Dict[Coord, int]]:
        """Get shortest path distance from every origin to all other coords."""
        src_dists = {}
        for origin in origins:
            src_dists[origin] = self.dijkstra(origin)
        return src_dists

    def dijkstra(self, origin: Coord) -> Dict[Coord, int]:
        """Get shortest path distance between origin and all other coords."""
        dist = {origin: 0}
        pq = PriorityQueue()  # type: ignore
        pq.put((dist[origin], origin))

        visited = {origin}

        while not pq.empty():
            _, coord = pq.get()

            for adj_coord in self.get_neighbours(coord, False):
                if dist[coord] + 1 < dist.get(adj_coord, float("inf")):
                    dist[adj_coord] = dist[coord] + 1

                    if adj_coord not in visited:
                        pq.put((dist[adj_coord], adj_coord))
                        visited.add(adj_coord)
        return dist

    def get_connected_components(self) -> List[Set[Coord]]:
        """Get list of connected components.

        A connected component is the set of all coords that are connected to
        each other - i.e. can be reached from each other.
        """
        components = []
        to_explore = set(self.unblocked_coords)

        while len(to_explore) > 0:
            coord = to_explore.pop()
            current_component = set({coord})
            to_expand: Queue[Coord] = Queue()
            to_expand.put(coord)

            while not to_expand.empty():
                coord = to_expand.get()
                for adj_coord in self.get_neighbours(coord):
                    if adj_coord in to_explore:
                        to_explore.remove(adj_coord)
                        to_expand.put(adj_coord)
                        current_component.add(adj_coord)

            components.append(current_component)
        return components

    def map_relative_to_absolute_coord(
        self,
        rel_coord: Coord,
        origin: Coord,
        facing_dir: Direction,
        forward_dim: int,
        side_dim: int,
    ) -> Coord:
        """Map a coord relative origin to coord on the grid.

        This is useful for mapping from coords in agent observations to coords in the
        actual grid.

        Arguments
        ---------
        rel_coord: the relative coordinate. This is the coordinate relative to the
            origin coordinate.
        origin: the origin coordinate. This is the coordinate in the actual grid that
            the relative coordinate is based on.
        facing_dir: the direction the relative grid is oriented.
        forward_dim: the distance from the origin in the facing dir that the relative
            grid reaches. Defines where the zeroth row of the relative grid is relative
            to the origin.
        side_dim: the distance from the origin perpendicular to the facing dir that the
            relative grid reaches. Defines where the zeroth column of the grid is
            relative to the origin.

        Returns
        -------
        The relative coord mapped to the actual grid.

        """
        grid_col, grid_row = origin

        if facing_dir == Direction.NORTH:
            grid_row += rel_coord[1] - forward_dim
            grid_col += rel_coord[0] - side_dim
        elif facing_dir == Direction.EAST:
            grid_row += rel_coord[0] - side_dim
            grid_col += -rel_coord[1] + forward_dim
        elif facing_dir == Direction.SOUTH:
            grid_row += -rel_coord[1] + forward_dim
            grid_col += -rel_coord[0] + side_dim
        else:
            grid_row += -rel_coord[0] + side_dim
            grid_col += rel_coord[1] - forward_dim
        return (grid_col, grid_row)

    def get_rectangular_bounds(
        self, origin: Coord, facing_dir: Direction, rect_size: Tuple[int, int, int, int]
    ) -> Tuple[int, int, int, int]:
        """Get rectangular bounds for a rectangle with given origin and size.

        Arguments
        ---------
        origin: the origin coordinate for the rectangle
        facing_dir: which direction the rectangle is oriented
        rect_size: the size of the rectangle relative to the origin. Specifically,
            rect_size[0] = number of cells in front
            rect_size[1] = number of cells behind
            rect_size[2] = number of cells left
            rect_size[3] = number of cells right

        Returns
        -------
        min_col, max_col, min_row, max_row of rectangle that is within the grid's
            bounds. Note that this may be smaller than the rectangle as specified by the
            rect_size.

        """
        assert 0 <= origin[0] < self.width and 0 <= origin[1] < self.height
        assert all(d >= 0 for d in rect_size)
        col, row = origin

        # rotate rectangle sizes based on direction so they are treated as if
        # direction was NORTH
        front, back, left, right = rect_size
        if facing_dir == Direction.SOUTH:
            front, back, left, right = back, front, right, left
        elif facing_dir == Direction.EAST:
            front, back, left, right = left, right, back, front
        elif facing_dir == Direction.WEST:
            front, back, left, right = right, left, front, back

        min_col = max(0, col - left)
        max_col = min(self.width - 1, col + right)
        min_row = max(0, row - front)
        max_row = min(self.height - 1, row + back)
        return min_col, max_col, min_row, max_row

    def get_rectangular_padding(
        self, origin: Coord, facing_dir: Direction, rect_size: Tuple[int, int, int, int]
    ) -> Tuple[Tuple[int, int], Tuple[int, int]]:
        """Get padding for a rectangle with given origin and size.

        The padding quantities are the number of cells the rectangle is out-of-bounds
        on each dimension.

        Arguments
        ---------
        origin: the origin coordinate for the rectangle
        facing_dir: which direction the rectangle is oriented
        rect_size: the size of the rectangle relative to the origin. Specifically,
            rect_size[0] = number of cells in front
            rect_size[1] = number of cells behind
            rect_size[2] = number of cells left
            rect_size[3] = number of cells right

        Returns
        -------
        (before_col, after_col), (before_row, after_row padding)

        """
        assert 0 <= origin[0] < self.width and 0 <= origin[1] < self.height
        assert all(d >= 0 for d in rect_size)
        col, row = origin

        # rotate rectangle sizes based on direction so they are treated as if
        # direction was NORTH
        front, back, left, right = rect_size
        if facing_dir == Direction.SOUTH:
            front, back, left, right = back, front, right, left
        elif facing_dir == Direction.EAST:
            front, back, left, right = left, right, back, front
        elif facing_dir == Direction.WEST:
            front, back, left, right = right, left, front, back

        before_col = -min(0, col - left)
        after_col = max(0, col + right - self.width + 1)
        before_row = -min(0, row - front)
        after_row = max(0, row + back - self.height + 1)
        return (before_col, after_col), (before_row, after_row)


class GridGenerator:
    """Class for generating grid layouts.

    Note, a block in the grid is an obstacle occupying a single cell, while
    here and obstacle refers to a collection of one or more connected blocks.
    """

    def __init__(
        self,
        width: int,
        height: int,
        mask: Set[Coord],
        max_obstacle_size: int,
        max_num_obstacles: int,
        ensure_grid_connected: bool,
        seed: Optional[int] = None,
    ):
        assert max_obstacle_size > 0
        self.width = width
        self.height = height
        self.mask = mask
        self.max_obstacle_size = max_obstacle_size
        self.max_num_obstacles = max_num_obstacles
        self.ensure_grid_connected = ensure_grid_connected
        self._rng = random.Random(seed)

    def generate(self) -> Grid:
        """Generate a new grid."""
        block_coords: Set[Coord] = set()
        for _ in range(self.max_num_obstacles):
            obstacle = self._get_random_obstacle()
            if not self.mask.intersection(obstacle):
                block_coords.update(obstacle)

        grid = Grid(self.width, self.height, block_coords)
        if self.ensure_grid_connected:
            grid = self.connect_grid_components(grid)

        return grid

    def generate_n(self, n: int) -> List[Grid]:
        """Generate N new grids."""
        grids = [self.generate() for _ in range(n)]
        return grids

    def _get_random_obstacle(self) -> Set[Coord]:
        obstacle_height = self._rng.randint(1, self.max_obstacle_size)
        obstacle_width = self._rng.randint(1, self.max_obstacle_size)
        obstacle_x = self._rng.randint(0, self.width - 1)
        obstacle_y = self._rng.randint(0, self.height - 1)
        obstacle = set()
        for x in range(obstacle_x, obstacle_x + obstacle_width):
            if x >= self.width:
                break
            for y in range(obstacle_y, obstacle_y + obstacle_height):
                if y >= self.height:
                    break
                obstacle.add((x, y))
        return obstacle

    def connect_grid_components(self, grid: Grid) -> Grid:
        """Get grid with only a single connected component.

        Removes blocks along shortest path between components until there is
        only a single connected component.
        """
        # make a copy of original grid
        grid = Grid(grid.width, grid.height, set(grid.block_coords))

        components = grid.get_connected_components()
        if len(components) == 1:
            return grid

        # connect components in order of distance from (0, 0)
        component_dists = [
            (self._component_distance(grid, (0, 0), c), i)
            for i, c in enumerate(components)
        ]
        component_dists.sort()

        single_component = components[component_dists[0][1]]
        for j in range(1, len(components)):
            component_j = components[component_dists[j][1]]
            closest_pair = self._get_closest_pair(grid, single_component, component_j)

            path = self._get_shortest_direct_path(grid, *closest_pair)
            for coord in path:
                if coord in grid.block_coords:
                    grid.block_coords.remove(coord)

            single_component.update(component_j)
            single_component.update(path)

        return grid

    def _component_distance(
        self, grid: Grid, origin: Coord, component: Set[Coord]
    ) -> int:
        return min([grid.manhattan_dist(origin, coord) for coord in component])

    def _get_closest_pair(
        self, grid: Grid, component_0: Set[Coord], component_1: Set[Coord]
    ) -> Tuple[Coord, Coord]:
        min_dist = grid.width * grid.height
        min_pair = ((0, 0), (0, 0))
        for c0, c1 in itertools.product(component_0, component_1):
            d = grid.manhattan_dist(c0, c1)
            if d < min_dist:
                min_dist = d
                min_pair = (c0, c1)
        return min_pair

    def _get_shortest_direct_path(
        self, grid: Grid, start_coord: Coord, goal_coord: Coord
    ) -> List[Coord]:
        """Get shortest direct path between two coords.

        This will possibly include blocks in the path, but will greedily chose
        paths that do not contain a block. There's no guarantee it'll be the
        direct path with the least number of blocks.
        """
        path = [start_coord]

        x_dir = 0
        if start_coord[0] != goal_coord[0]:
            x_dir = 1 if start_coord[0] < goal_coord[0] else -1

        y_dir = 0
        if start_coord[1] != goal_coord[1]:
            y_dir = 1 if start_coord[1] < goal_coord[1] else -1

        coord = start_coord
        while coord != goal_coord:
            next_coords = []
            if coord[0] != goal_coord[0]:
                next_coords.append((coord[0] + x_dir, coord[1]))
            if coord[1] != goal_coord[1]:
                next_coords.append((coord[0], coord[1] + y_dir))

            if len(next_coords) == 1:
                coord = next_coords[0]
                path.append(coord)
                continue

            blocked = [c in grid.block_coords for c in next_coords]
            if all(blocked) or not any(blocked):
                coord = self._rng.choice(next_coords)
            else:
                # if [0] is blocked, choose [1], and vice versa
                coord = next_coords[blocked[0]]
            path.append(coord)

        return path

    def get_grid_str(self, grid: Grid) -> str:
        """Get a str representation of a grid with mask applied."""
        grid_repr = []
        for row in range(grid.height):
            row_repr = []
            for col in range(grid.width):
                coord = (col, row)
                if coord in grid.block_coords:
                    row_repr.append("#")
                elif coord in self.mask:
                    row_repr.append("M")
                else:
                    row_repr.append(".")
            grid_repr.append(row_repr)

        return "\n".join([" ".join(r) for r in grid_repr])


class GridCycler:
    """Class for handling cycling through a set of generated grids."""

    def __init__(
        self, grids: List[Grid], shuffle_each_cycle: bool, seed: Optional[int] = None
    ):
        self.grids = grids
        self.shuffle = shuffle_each_cycle
        self._rng = random.Random(seed)

        self._next_idx = 0

    def next(self) -> Grid:
        if self._next_idx >= len(self.grids):
            self._next_idx = 0

            if self.shuffle:
                self._rng.shuffle(self.grids)

        grid = self.grids[self._next_idx]
        self._next_idx += 1
        return grid



"""Wrapper to rescale continuous actions from [min, max] range."""
from typing import Dict, Union

import numpy as np
from gymnasium import spaces

import posggym


class RescaleActions(posggym.ActionWrapper):
    """Action wrapper that rescales continuous action space.

    Rescales actions from [min, max] range to the action space range of the environment.

    The base environment :attr:`env` must have an action space of type
    :class:`spaces.Box` for each agent. If :attr:`min_action` or :attr:`max_action` are
    numpy arrays, the shape must match the shape of the environment's action space for
    the given agent. If :attr:`min_action` or :attr:`max_action` are dictionaries then
    they must have an entry for each possible agent ID in the wrapped environment.

    Arguments
    ---------
    env : posggym.Env
        The environment to apply the wrapper
    min_action : float, int, np.ndarray, Dict[str, Union[float, int, np.ndarray]]
        The minimum value for the scaled actions.
    max_action : float, int, np.ndarray, Dict[str, Union[float, int, np.ndarray]]
        The maximum value for the scaled actions.

    Note
    ----
    Explanation of how to scale number from one interval into new interval:
    https://stats.stackexchange.com/questions/281162/scale-a-number-between-a-range

    Implementation also based on similar wrapper from gymnasium:
    https://github.com/Farama-Foundation/Gymnasium/blob/v0.27.0/gymnasium/wrappers/rescale_action.py

    """

    def __init__(
        self,
        env: posggym.Env,
        min_action: Union[
            float, int, np.ndarray, Dict[str, Union[float, int, np.ndarray]]
        ],
        max_action: Union[
            float, int, np.ndarray, Dict[str, Union[float, int, np.ndarray]]
        ],
    ):
        self.min_action = {}
        self.max_action = {}
        self.rescale_factor = {}
        for i, action_space in env.action_spaces.items():
            assert isinstance(action_space, spaces.Box), (
                f"Expected Box action space for all agents, got {type(action_space)} "
                f"for agent '{i}'."
            )

            min_action_i = min_action[i] if isinstance(min_action, dict) else min_action
            if isinstance(min_action_i, (float, int)):
                self.min_action[i] = np.full_like(action_space.low, min_action_i)
            else:
                assert isinstance(min_action_i, np.ndarray), min_action_i
                assert min_action_i.shape == action_space.shape, (
                    min_action_i.shape,
                    action_space.shape,
                )
                self.min_action[i] = min_action_i

            max_action_i = max_action[i] if isinstance(max_action, dict) else max_action
            if isinstance(max_action_i, (float, int)):
                self.max_action[i] = np.full_like(action_space.high, max_action_i)
            else:
                assert isinstance(max_action_i, np.ndarray), max_action_i
                assert max_action_i.shape == action_space.shape, (
                    max_action_i.shape,
                    action_space.shape,
                )
                self.max_action[i] = max_action_i

            assert np.less_equal(self.min_action[i], self.max_action[i]).all(), (
                self.min_action[i],
                self.max_action[i],
            )
            self.rescale_factor[i] = (action_space.high - action_space.low) / (
                self.max_action[i] - self.min_action[i]
            )

        super().__init__(env)
        self._action_spaces = {
            i: spaces.Box(low=self.min_action[i], high=self.max_action[i])
            for i in env.action_spaces
        }

    def actions(self, actions):
        # rescales from [min_action, max_action] to the action space of wrapped env.
        rescaled_actions = {}
        for i, act_i in actions.items():
            low = self.env.action_spaces[i].low
            high = self.env.action_spaces[i].high
            rescaled_action = (act_i - self.min_action[i]) * self.rescale_factor[
                i
            ] + low
            rescaled_actions[i] = np.clip(rescaled_action, low, high)
        return rescaled_actions



"""Policies for the DrivingContinuous-v0 environment."""
import os.path as osp

from posggym.agents.torch_policy import PPOPolicy
from posggym.agents.utils import processors
from posggym.config import AGENT_MODEL_DIR

ENV_ID = "DrivingContinuous-v0"
agent_model_dir = osp.join(AGENT_MODEL_DIR, "continuous", "driving_continuous")
policy_specs = {}

# 14x14RoundAbout, 2 agents
for policy_file_name in [
    "sp_seed0.pkl",
    "sp_seed1.pkl",
    "sp_seed2.pkl",
    "sp_seed3.pkl",
    "sp_seed4.pkl",
]:
    spec = PPOPolicy.get_spec_from_path(
        env_id=ENV_ID,
        env_args={
            "world": "14x14RoundAbout",
            "num_agents": 2,
            "obs_dist": 5.0,
            "n_sensors": 16,
        },
        policy_file_path=osp.join(
            agent_model_dir,
            "world=14x14RoundAbout-num_agents=2-obs_dist=5-n_sensors=16",
            policy_file_name,
        ),
        version=0,
        valid_agent_ids=None,
        # policy is deterministic given random seed
        nondeterministic=False,
        # actions sampled, rather than always taking most probable action
        deterministic=False,
        obs_processor_cls=processors.RescaleProcessor,
        obs_processor_config={"min_val": -1, "max_val": 1},
        action_processor_cls=processors.RescaleProcessor,
        action_processor_config={"min_val": -1, "max_val": 1, "clip": True},
    )
    policy_specs[spec.id] = spec


# 14x14RoundAbout, 4 agents
for policy_file_name in [
    "sp_seed0.pkl",
    "sp_seed1.pkl",
    "sp_seed2.pkl",
    "sp_seed3.pkl",
    "sp_seed4.pkl",
]:
    spec = PPOPolicy.get_spec_from_path(
        env_id=ENV_ID,
        env_args={
            "world": "14x14RoundAbout",
            "num_agents": 4,
            "obs_dist": 5.0,
            "n_sensors": 16,
        },
        policy_file_path=osp.join(
            agent_model_dir,
            "world=14x14RoundAbout-num_agents=4-obs_dist=5-n_sensors=16",
            policy_file_name,
        ),
        version=0,
        valid_agent_ids=None,
        # policy is deterministic given random seed
        nondeterministic=False,
        # actions sampled, rather than always taking most probable action
        deterministic=False,
        obs_processor_cls=processors.RescaleProcessor,
        obs_processor_config={"min_val": -1, "max_val": 1},
        action_processor_cls=processors.RescaleProcessor,
        action_processor_config={"min_val": -1, "max_val": 1, "clip": True},
    )
    policy_specs[spec.id] = spec



"""The Drone Team Capture Environment."""
import math
from typing import Dict, List, NamedTuple, Optional, Tuple, cast

import numpy as np
from gymnasium import spaces

import posggym.model as M
from posggym import logger
from posggym.core import DefaultEnv
from posggym.envs.continuous.core import (
    CircularContinuousWorld,
    PMBodyState,
    clip_actions,
)
from posggym.utils import seeding


class DTCState(NamedTuple):
    """A state in the Drone Team Capture Environment."""

    pursuer_states: np.ndarray
    prev_pursuer_states: np.ndarray
    target_state: np.ndarray
    prev_target_state: np.ndarray
    target_vel: float


DTCAction = np.ndarray
DTCObs = np.ndarray


class DroneTeamCaptureEnv(DefaultEnv[DTCState, DTCObs, DTCAction]):
    """The Drone Team Capture Environment.

    A co-operative 2D continuous world problem involving multiple pursuer drone agents
    working together to catch a target agent in the environment.

    This is an adaption of the original code base to add partially observability.
    Specifically, it is possible to limit the observation range of the pursuers
    (see `observation_limit` argument) and also how many other pursuers each pursuer
    can observe at a time, i.e.how many in range pursuers each pursuer can communicate
    with (see `n_communicating_pursuers` argument). Furthermore we extent the
    observation space of each agent compared to the original code by making it so each
    pursuer also observes their own (x, y) position.

    By default, all pursuers can observe their own angle/yaw and angular velocity,
    as well as the relative position of the target and all other pursuers.

    Possible Agents
    ---------------
    Varied number (1-8)

    State Space
    -----------
    Each state consists of:

    1. 2D array containing the state of each pursuer
    2. 2D array containing the previous state of each pursuer
    3. 1D array containing the state of the target
    4. 1D array containing the previous state of the target
    5. The velocity of the target (between 0.5 and 2.0 time max velocity of pursuers)

    The state of each pursuer and the target are a 1D array containing their:

    - `(x, y)` coordinate
    - angle/yaw (in radians)
    - velocity in x and y directions
    - angular velocity (in radians)

    Action Space
    ------------
    Each agent has either 1 or 2 actions. If 'velocity_control=False' then the agent
    can only control their angular velocity. If 'velocity_control=True' then the agent
    has two actions, which are the angular and linear velocity, in that order.

    Observation Space
    -----------------
    Each agent receives a 1D vector observation containing information about their
    current state as well as some information about the other pursuers and the target.

    - *Self obs* - observe angle that pursuer is facing, current angular velocity, and
        current (x, y) position.
    - *Target obs* - observe the angle and distance to the target (if target is within
        `observation_limit`), as well as rate of change of angle and distance to the
        target.
    - *Other pursuer obs* - observes the angle and distance to the
        `n_communicating_pursuers` closest pursuer agents (if they are within
        `observation_limit` distance).

    All observation features have values normalized into [-1, 1] interval. Observation
    feature values are `-1` for any pursuers or the target if they are out of range
    observation range.

    This table enumerates the observation space:

    | Index: start          | Description                          |  Values   |
    | :-------------------: | :----------------------------------: | :-------: |
    | 0                     | Agent angle                          | [-1, 1[   |
    | 1                     | Agent angular velocity               | [-1, 1]   |
    | 2                     | Agent (x, y) position                | [-1, 1]   |
    | 4                     | Angle to target                      | [-1, 1]   |
    | 5                     | Distance to target                   | [-1, 1]   |
    | 6                     | Angular velocity of angle to target  | [-1, 1]   |
    | 7                     | Velocity of distance to target       | [-1, 1]   |
    | 8 to (8 + 2 * n)      | Other pursuer angle and distance     | [-1, 1]   |

    Where `n = n_communicating_pursuers`.

    Rewards
    -------
    Each pursuer will receive a capture reward and a reward based on their distance from
    the target. On successful capture, the capturing pursuer will receive a reward of
    `+130`, while other agents will receive `100`.

    Optionally, the pursuers receive a reward based on the `Q` parameter which measures
    the spread of the pursuers and incentivizes them to spread out around the target.

    Dynamics
    --------
    Actions of the pursuer agents are deterministic and consist of moving based on the
    angular and linear velocity.

    The target has a maximum velocity that varies per episodes and it actively moves
    away from the pursuers and also the walls.

    Starting State
    --------------
    Target will start near the outside of the circle, while pursuers will start in a
    line on the middle.

    The velocity of the target is chosen uniformly at random at the start of each
    episode to be between 0.5 and 2.0 times the max velocity of the pursuers.

    Episodes End
    ------------
    Episodes ends when the target has been captured. By default a `max_episode_steps`
    limit of `100` steps is also set. This may need to be adjusted when using larger
    world sizes (this can be done by manually specifying a value for
    `max_episode_steps` when creating the environment with `posggym.make`).


    Arguments
    ---------

    - `num_agents` - The number of agents which exist in the environment
        Must be between 1 and 8 (default = `3`)
    - `n_communicating_pursuers - The maximum number of agents which an
        agent can receive information from. If `None` then this will be set to equal
        `num_agents - 1` (default = `None`)
    - `arena_radius` - Size of the arena, in terms of it's radius (default = `430`)
    - `observation_limit` - The limit of which agents can see other agents, if `None`
        then there is no observation limit (default = `None`)
    - `velocity_control` - If the agents have control of their linear velocity
        (default = `False`)
    - `capture_radius` - Distance from target pursuer needs to be within to capture
        target. As per original paper, the user can adjust this to set a learning
        curriculum (larger values are easier) (default = `30`, which is the radius of
        the target).
    - `use_q_reward` - Whether the pursuers should also receive the reward based on the
        `Q` parameter (default = `False`)


    Available variants
    ------------------
    For example to use the Drone Team Capture environment with 8 pursuer drones, with
    communication between max 4 closest drones and episode step limit of 100, and the
    default values for the other parameters (`velocity_control`, `arena_radius`,
    `observation_limit`, `capture_radius`) you would use:

    ```python
    import posggym
    env = posggym.make(
        'DroneTeamCapture-v0',
        max_episode_steps=100,
        num_agents=8,
        n_communicating_pursuers=4,
    )
    ```

    Version History
    ---------------
    - `v0`: Initial version

    Reference
    ---------
    - C. de Souza, R. Newbury, A. Cosgun, P. Castillo, B. Vidolov and D. Kulić,
      "Decentralized Multi-Agent Pursuit Using Deep Reinforcement Learning,"
      in IEEE Robotics and Automation Letters, vol. 6, no. 3, pp. 4552-4559,
      July 2021, doi: 10.1109/LRA.2021.3068952.

    """

    metadata = {
        "render_modes": ["human", "rgb_array"],
        "render_fps": 15,
    }

    def __init__(
        self,
        num_agents: int = 3,
        n_communicating_pursuers: Optional[int] = None,
        arena_radius: float = 430,
        observation_limit: Optional[float] = None,
        velocity_control: bool = False,
        capture_radius: float = 30,
        use_q_reward: bool = False,
        render_mode: Optional[str] = None,
    ):
        super().__init__(
            DroneTeamCaptureModel(
                num_agents,
                n_communicating_pursuers=n_communicating_pursuers,
                arena_radius=arena_radius,
                observation_limit=observation_limit,
                velocity_control=velocity_control,
                capture_radius=capture_radius,
                use_q_reward=use_q_reward,
            ),
            render_mode=render_mode,
        )

        self._viewer = None
        self.window_surface = None
        self._renderer = None
        self.render_mode = render_mode
        self.clock = None

        self.window_size = 600
        self.draw_options = None
        self.world = None

    def render(self):
        if self.render_mode is None:
            assert self.spec is not None
            logger.warn(
                "You are calling render method without specifying any render mode. "
                "You can specify the render_mode at initialization, "
                f'e.g. posggym.make("{self.spec.id}", render_mode="rgb_array")'
            )
            return
        return self._render_img()

    def _render_img(self):
        import pygame
        from pymunk import Transform, pygame_util

        model = cast(DroneTeamCaptureModel, self.model)
        state = cast(DTCState, self.state)
        scale_factor = self.window_size / model.world.size

        if self.window_surface is None:
            pygame.init()
            if self.render_mode == "human":
                pygame.display.init()
                pygame.display.set_caption(self.__class__.__name__)
                self.window_surface = pygame.display.set_mode(
                    (self.window_size, self.window_size)
                )
            else:
                self.window_surface = pygame.Surface(
                    (self.window_size, self.window_size)
                )
            # Turn off alpha since we don't use it.
            self.window_surface.set_alpha(None)

        if self.clock is None:
            self.clock = pygame.time.Clock()

        if self.draw_options is None:
            pygame_util.positive_y_is_up = False
            self.draw_options = pygame_util.DrawOptions(self.window_surface)
            self.draw_options.transform = Transform.scaling(scale_factor)
            # don't render collision lines
            self.draw_options.flags = (
                pygame_util.DrawOptions.DRAW_SHAPES
                | pygame_util.DrawOptions.DRAW_CONSTRAINTS
            )

        if self.world is None:
            # get copy of model world, so we can use it for rendering without
            # affecting the original
            self.world = model.world.copy()

        for i, p_state in enumerate(state.pursuer_states):
            self.world.set_entity_state(f"pursuer_{i}", p_state)

        self.world.set_entity_state("evader", state.target_state)

        # Need to do this for space to update with changes
        self.world.space.step(0.0001)

        # reset screen
        self.window_surface.fill(pygame.Color("white"))

        self.world.space.debug_draw(self.draw_options)

        if self.render_mode == "human":
            pygame.event.pump()
            pygame.display.update()
            self.clock.tick(self.metadata["render_fps"])
            return None

        return np.transpose(
            np.array(pygame.surfarray.pixels3d(self.window_surface)), axes=(1, 0, 2)
        )


class DroneTeamCaptureModel(M.POSGModel[DTCState, DTCObs, DTCAction]):
    """The Drone Team Capture problem model.

    Parameters
    ----------
    num_agents: int
        The number of agents which exist in the environment. Must be between 1 and 8
    n_communicating_pursuers: int, optional
        The maximum number of agents which an agent can receive information from. If
        None, then all agents can communicate with each other.
    arena_radius : float
        Size of the arena, in terms of its radius
    observation_limit : float, optional
        The limit of which agents can see other agents. If None, then there is no
        observation limit.
    velocity_control: bool
        If the agents have control of their linear velocity
    capture_radius : float
        Distance from target pursuer needs to be within to capture target. As per
        original paper, the user can adjust this to set a learning curriculum (larger
        values are easier).
    use_q_reward : bool
        Whether to include the q-formation reward in the reward function

    """

    # reward received by all pursuers if target is captured
    R_CAPTURE_TEAM = 100
    # additional reward received by any capturing pursuer (30% more than the others)
    R_CAPTURE = 30
    # Coefficient for q-formation reward in reward function
    R_Q_COEFF = -0.1
    # Coefficient for distance to target in reward function
    R_TARGET_DIST_COEFF = -0.002

    PURSUER_COLOR = (55, 155, 205, 255)  # blueish
    EVADER_COLOR = (110, 55, 155, 255)  # purpleish

    def __init__(
        self,
        num_agents: int,
        n_communicating_pursuers: Optional[int] = None,
        arena_radius: float = 430,
        observation_limit: Optional[float] = None,
        velocity_control: bool = False,
        capture_radius: float = 30,
        use_q_reward: bool = False,
    ):
        assert 1 < num_agents <= 8
        assert (
            n_communicating_pursuers is None
            or 0 < n_communicating_pursuers < num_agents
        )
        self.n_pursuers = num_agents
        self.n_com_pursuers = (
            n_communicating_pursuers if n_communicating_pursuers else num_agents - 1
        )
        self.velocity_control = velocity_control
        self.r_arena = arena_radius
        self.observation_limit = observation_limit
        self.capture_radius = capture_radius
        self.use_q_reward = use_q_reward

        # Fixed model params
        # Linear Velocity of pursuer
        self.max_pursuer_vel = 10
        # Min and Max relative linear velocity of target
        self.target_rel_vel_bounds = (0.5, 2.0)
        # max relative change in position between any two entities
        self.max_rel_dist_change = self.max_pursuer_vel * (
            1 + self.target_rel_vel_bounds[1]
        )
        self.norm_max_rel_dist_change = self.max_rel_dist_change / (2 * self.r_arena)

        self.possible_agents = tuple(str(x) for x in range(self.n_pursuers))
        self.is_symmetric = True

        self.dyaw_limit = math.pi / 10
        if self.velocity_control:
            # act[0] = angular velocity, act[1] = linear velocity
            self.action_spaces = {
                i: spaces.Box(
                    low=np.array([-self.dyaw_limit, 0], dtype=np.float32),
                    high=np.array([self.dyaw_limit, 1], dtype=np.float32),
                )
                for i in self.possible_agents
            }
        else:
            # act[0] = angular velocity
            self.action_spaces = {
                i: spaces.Box(
                    low=np.array([-self.dyaw_limit], dtype=np.float32),
                    high=np.array([self.dyaw_limit], dtype=np.float32),
                )
                for i in self.possible_agents
            }

        self.obs_dim = 8 + self.n_com_pursuers * 2
        # standard range for raw observation values (before normalization to [-1, 1])
        self.raw_obs_range = (
            np.array(
                [
                    -math.pi,
                    -self.dyaw_limit,
                    0.0,
                    0.0,
                    -math.pi,
                    0,
                    -self.dyaw_limit,
                    -self.max_rel_dist_change,
                ]
                + [
                    -math.pi if i % 2 == 0 else 0
                    for i in range(self.n_com_pursuers * 2)
                ],
                dtype=np.float32,
            ),
            np.array(
                [
                    math.pi,
                    self.dyaw_limit,
                    2 * self.r_arena,
                    2 * self.r_arena,
                    math.pi,
                    2 * self.r_arena,
                    self.dyaw_limit,
                    self.max_rel_dist_change,
                ]
                + [
                    math.pi if i % 2 == 0 else 2 * self.r_arena
                    for i in range(self.n_com_pursuers * 2)
                ],
                dtype=np.float32,
            ),
            np.full((self.obs_dim,), 2 * self.r_arena, dtype=np.float32),
        )
        self.observation_spaces = {
            i: spaces.Box(low=-1.0, high=1.0, shape=(self.obs_dim,), dtype=np.float32)
            for i in self.possible_agents
        }

        self.world = CircularContinuousWorld(
            size=self.r_arena * 2,
            agent_radius=30.0,
            blocks=None,
            enable_agent_collisions=False,
            border_thickness=2,
        )

        # Add physical entities to the world
        for i in range(self.n_pursuers):
            self.world.add_entity(f"pursuer_{i}", None, color=self.PURSUER_COLOR)
        self.world.add_entity("evader", None, color=self.EVADER_COLOR)

    def get_agents(self, state: DTCState) -> List[str]:
        return list(self.possible_agents)

    @property
    def reward_ranges(self) -> Dict[str, Tuple[float, float]]:
        min_reward = self.R_TARGET_DIST_COEFF * 2 * self.r_arena
        max_reward = self.R_CAPTURE_TEAM + self.R_CAPTURE
        if self.use_q_reward:
            n = self.n_pursuers
            min_reward += self.R_Q_COEFF * 3.0 * (n - 1) / n
            max_reward += self.R_Q_COEFF * -1.0 * (n - 1) / n

        return {i: (min_reward, max_reward) for i in self.possible_agents}

    @property
    def rng(self) -> seeding.RNG:
        if self._rng is None:
            self._rng, _ = seeding.std_random()
        return self._rng

    def sample_initial_state(self) -> DTCState:
        pursuer_states = np.zeros(
            (self.n_pursuers, PMBodyState.num_features()), dtype=np.float32
        )
        for i in range(self.n_pursuers):
            # distributes the agents based on their index
            x = 50.0 * (-math.floor(self.n_pursuers / 2) + i) + self.r_arena
            pursuer_states[i][:3] = (x, self.r_arena, 0.0)

        # Target is placed randomly in sphere,
        # excluding area near center where pursuers start
        # ref: https://stackoverflow.com/questions/5837572/generate-a-random-point-within-a-circle-uniformly
        min_r = 4 * self.world.agent_radius
        max_r = self.r_arena - self.world.agent_radius
        r = math.sqrt(self.rng.random()) * (max_r - min_r) + min_r
        theta = self.rng.random() * 2 * math.pi
        x = self.r_arena + r * math.cos(theta)
        y = self.r_arena + r * math.sin(theta)

        # The velocity of the target varies
        min_rvel, max_rvel = self.target_rel_vel_bounds
        relative_target_vel = float(
            self.rng.random() * (max_rvel - min_rvel) + min_rvel
        )
        target_vel = relative_target_vel * self.max_pursuer_vel

        target_state = np.zeros((PMBodyState.num_features()), dtype=np.float32)
        target_state[:3] = (x, y, 0.0)

        return DTCState(
            pursuer_states,
            np.copy(pursuer_states),
            target_state,
            np.copy(target_state),
            target_vel,
        )

    def sample_initial_obs(self, state: DTCState) -> Dict[str, DTCObs]:
        return self._get_obs(state)

    def step(
        self, state: DTCState, actions: Dict[str, DTCAction]
    ) -> M.JointTimestep[DTCState, DTCObs]:
        clipped_actions = clip_actions(actions, self.action_spaces)
        next_state = self._get_next_state(state, clipped_actions)
        obs = self._get_obs(next_state)
        all_done, rewards = self._get_rewards(next_state)
        terminations = {i: all_done for i in self.possible_agents}
        truncations = {i: False for i in self.possible_agents}
        infos: Dict[str, Dict] = {i: {} for i in self.possible_agents}
        return M.JointTimestep(
            next_state, obs, rewards, terminations, truncations, all_done, infos
        )

    def _get_next_state(
        self, state: DTCState, actions: Dict[str, DTCAction]
    ) -> DTCState:
        for i in range(self.n_pursuers):
            self.world.set_entity_state(f"pursuer_{i}", state.pursuer_states[i])
            velocity_factor = 1 if not self.velocity_control else actions[str(i)][1]
            pursuer_angle = state.pursuer_states[i][2] + actions[str(i)][0]
            pursuer_vel = self.world.linear_to_xy_velocity(
                velocity_factor * self.max_pursuer_vel, pursuer_angle
            )
            self.world.update_entity_state(
                f"pursuer_{i}",
                angle=pursuer_angle,
                vel=pursuer_vel,
            )

        evader_vel_xy = self._get_target_move_repulsive(state)
        self.world.set_entity_state("evader", state.target_state)
        self.world.update_entity_state("evader", vel=evader_vel_xy)

        self.world.simulate(1.0 / 10, 10, normalize_angles=True)

        next_pursuer_states = np.array(
            [
                self.world.get_entity_state(f"pursuer_{i}")
                for i in range(self.n_pursuers)
            ],
            dtype=np.float32,
        )
        return DTCState(
            next_pursuer_states,
            np.copy(state.pursuer_states),
            np.array(self.world.get_entity_state("evader"), dtype=np.float32),
            np.copy(state.target_state),
            state.target_vel,
        )

    def _get_obs(self, state: DTCState) -> Dict[str, DTCObs]:
        observation = {}
        for i in range(self.n_pursuers):
            # getting the target engagement
            (alpha_t, dist_t), target_visible = self._engagement(
                state.pursuer_states[i],
                state.target_state,
                dist_norm_factor=2 * self.r_arena,
            )
            (alpha_t_prev, dist_t_prev), target_prev_visible = self._engagement(
                state.prev_pursuer_states[i],
                state.prev_target_state,
                dist_norm_factor=2 * self.r_arena,
            )
            # change in alpha
            if not target_visible or not target_prev_visible:
                alpha_rate = -1.0
                dist_rate = -1.0
            else:
                # alpha_t and alpha_t_prev are both normalized into [-1, 1] range
                # so have to do some shenanigans to ensure alpha rate is correctly
                # normalized into [-1, 1]
                alpha_rate = (
                    self.world.convert_angle_to_negpi_pi_interval(
                        (alpha_t - alpha_t_prev) * math.pi
                    )
                    / math.pi
                )
                max_rate = self.norm_max_rel_dist_change
                dist_rate = self.world.convert_into_interval(
                    dist_t - dist_t_prev, -max_rate, max_rate, -1.0, 1.0
                )

            # getting the relative engagement
            # alpha and distance to each pursuer
            engagement = []
            for j in range(self.n_pursuers):
                if j == i:
                    eng = (-1.0, -1.0)
                else:
                    eng, _ = self._engagement(
                        state.pursuer_states[i],
                        state.pursuer_states[j],
                        dist_norm_factor=2 * self.r_arena,
                    )
                engagement.append(eng)

            # Put any invalid (-1) to the end
            engagement = sorted(
                engagement, key=lambda t: float("inf") if t[1] == -1.0 else t[1]
            )
            alphas, dists = list(zip(*engagement))

            angle_i = (
                self.world.convert_angle_to_negpi_pi_interval(
                    state.pursuer_states[i][2]
                )
                / math.pi
            )
            prev_angle_i = (
                self.world.convert_angle_to_negpi_pi_interval(
                    state.prev_pursuer_states[i][2]
                )
                / math.pi
            )
            turn_rate = (angle_i - prev_angle_i) / 2
            xy_i = self.world.convert_into_interval(
                state.pursuer_states[i][:2], 0.0, 2 * self.r_arena, -1.0, 1.0
            )

            # Create obs vector
            obs_i = [
                angle_i,
                turn_rate,
                xy_i[0],
                xy_i[1],
                alpha_t,
                dist_t,
                alpha_rate,
                dist_rate,
            ]

            for idx in range(self.n_com_pursuers):
                obs_i.append(alphas[idx])
                obs_i.append(dists[idx])

            observation[str(i)] = np.array(obs_i, dtype=np.float32)

        return observation

    def _engagement(
        self, agent_i: np.ndarray, agent_j: np.ndarray, dist_norm_factor: float
    ) -> Tuple[Tuple[float, float], bool]:
        """Get engagement between two agents.

        Engagement here is the angle (in radians) from agent_i's current position and
        angle to agent_j's position, as well as the distance between the two agents
        positions. Both angle and distance are normalized to [-1, 1] range.

        Note, if agents are outside of observation distance of each other then returns
        (-1, -1), and False.

        """
        dist = self.world.euclidean_dist(agent_i, agent_j)
        if self.observation_limit is not None and dist > self.observation_limit:
            return (-1.0, -1.0), False

        # Rotation matrix of yaw
        yaw = agent_i[2]
        rot = np.array(
            [[math.cos(yaw), math.sin(yaw)], [-math.sin(yaw), math.cos(yaw)]]
        )
        rel_xy = agent_j[:2] - agent_i[:2]
        rel_xy = rot.dot(rel_xy)
        alpha = math.atan2(rel_xy[1], rel_xy[0])
        alpha = self.world.convert_angle_to_negpi_pi_interval(alpha)
        return (alpha / math.pi, dist / dist_norm_factor), True

    def _get_rewards(self, state: DTCState) -> Tuple[bool, Dict[str, float]]:
        done = False
        reward: Dict[str, float] = {}
        # q_formation reward: [-1 * (n-1) / n, 3 * (n-1) / n]
        q_formation = self._q_parameter(state) if self.use_q_reward else 0.0
        for i in self.possible_agents:
            reward[i] = self.R_Q_COEFF * q_formation
            # target_dist range = (0.0, 2*r_arena) = (0.0, 860) for default size
            target_dist = self._target_distance(state, int(i))
            # target_dist reward = (-1.72, 0.0) for default size
            reward[i] += self.R_TARGET_DIST_COEFF * target_dist
            if target_dist < self.capture_radius:
                done = True
                reward[i] += self.R_CAPTURE

        if done:
            # Large possible reward when done!
            for i in self.possible_agents:
                reward[i] += self.R_CAPTURE_TEAM

        return done, reward

    def _q_parameter(self, state: DTCState) -> float:
        """Calculate Q-formation value."""
        # min = -1 * (n-1) / n, max = 3 * (n-1) / n
        closest = self._get_closest_pursuer(state)
        unit = self._get_unit_vectors(state)
        Qk = 0.0
        for i in range(self.n_pursuers):
            if i != closest:
                Qk += np.dot(unit[i], unit[closest]) + 1.0
        Qk /= self.n_pursuers
        return Qk

    def _get_closest_pursuer(self, state: DTCState) -> int:
        """Get closest pursuer to target."""
        min_dist, min_index = None, None
        for idx in range(len(state.pursuer_states)):
            dist = self._target_distance(state, idx)
            if min_dist is None or dist < min_dist:
                min_index = idx
                min_dist = dist
        if min_dist is None or min_index is None:
            raise Exception("No closest index found. Something has gone wrong.")
        return min_index

    def _get_unit_vectors(self, state: DTCState) -> List[List[float]]:
        """Get unit vectors between target and each pursuer."""
        unit = []
        q = state.target_state[:2]
        for p in state.pursuer_states:
            dist = self.world.euclidean_dist(p, state.target_state)
            unit.append([(q[0] - p[0]) / dist, (q[1] - p[1]) / dist])
        return unit

    def _get_target_move_repulsive(self, state: DTCState) -> Tuple[float, float]:
        xy_pos = state.target_state[:2]
        x, y = xy_pos

        def scale_fn(z):
            return 50000 / (abs(z) + 200) ** 2

        final_vector = [0.0, 0.0]
        for s in state.pursuer_states:
            vector = s[:2] - xy_pos
            final_vector = self._scale_vector(vector, scale_fn, final_vector)

        # Find closest point on border then put it in to the vectorial sum
        # Noting coords are with origin at top left, so must translate to where origin
        # is center of circle to get closest point on border, then translate back before
        # adding to vectorial sum
        gamma = -math.atan2(y - self.r_arena, x - self.r_arena)
        virtual_wall = [
            self.r_arena + self.r_arena * math.cos(gamma),
            self.r_arena - self.r_arena * math.sin(gamma),
        ]
        vector = virtual_wall - xy_pos
        final_vector = self._scale_vector(
            vector, scale_fn, final_vector, 0.5 * len(state.pursuer_states)
        )

        scaled_move_dir = [v / self._abs_sum(final_vector) for v in final_vector]
        dx, dy = scaled_move_dir[0], scaled_move_dir[1]
        d = np.linalg.norm([dx, dy])
        dx = float(state.target_vel * dx / d)
        dy = float(state.target_vel * dy / d)
        return dx, dy

    def _scale_vector(self, vector, scale_fn, final_vector, factor=1.0) -> List[float]:
        vec_sum = self._abs_sum(vector)
        div = max(0.00001, vec_sum)
        f = -factor * scale_fn(vec_sum) / div
        vector = f * vector
        return [x + y for x, y in zip(final_vector, vector)]

    def _abs_sum(self, vector: List[float]) -> float:
        return sum([abs(x) for x in vector])

    def _target_distance(self, state: DTCState, pursuer_idx: int) -> float:
        return self.world.euclidean_dist(
            state.pursuer_states[pursuer_idx], state.target_state
        )



"""The Driving Grid World Environment."""
import enum
from itertools import product
from typing import Any, Dict, List, NamedTuple, Optional, Sequence, Set, Tuple, Union

from gymnasium import spaces

import posggym.model as M
from posggym import logger
from posggym.core import DefaultEnv
from posggym.envs.grid_world.core import DIRECTION_ASCII_REPR, Coord, Direction, Grid
from posggym.utils import seeding


class Speed(enum.IntEnum):
    """A speed setting for a vehicle."""

    REVERSE = 0
    STOPPED = 1
    FORWARD_SLOW = 2
    FORWARD_FAST = 3


class VehicleState(NamedTuple):
    """The state of a vehicle in the Driving Environment."""

    coord: Coord
    facing_dir: Direction
    speed: Speed
    dest_coord: Coord
    dest_reached: int
    crashed: int
    min_dest_dist: int
    init_dest_dist: int


DState = Tuple[VehicleState, ...]

# Initial direction and speed of each vehicle
INIT_DIR = Direction.NORTH
INIT_SPEED = Speed.STOPPED

# The actions
DAction = int
DO_NOTHING = 0
ACCELERATE = 1
DECELERATE = 2
TURN_RIGHT = 3
TURN_LEFT = 4

ACTIONS = [DO_NOTHING, ACCELERATE, DECELERATE, TURN_RIGHT, TURN_LEFT]
ACTIONS_STR = ["0", "acc", "dec", "tr", "tl"]

# Obs = [
# V0 Obs = (adj_obs, speed, dest_coord, dest_reached, crashed)
# V1 Obs = (adj_obs, speed, Coord, dest_coord, dest_reached, crashed)
DObs = Union[
    Tuple[Tuple[int, ...], Speed, Coord, int, int],
    Tuple[Tuple[int, ...], Speed, Coord, Coord, int, int],
]

# Cell obs
VEHICLE = 0
WALL = 1
EMPTY = 2
DESTINATION = 3

CELL_OBS = [VEHICLE, WALL, EMPTY, DESTINATION]
CELL_OBS_STR = ["V", "#", "0", "D"]


class DrivingEnv(DefaultEnv[DState, DObs, DAction]):
    """The Driving Grid World Environment.

    A general-sum 2D grid world problem involving multiple agents. Each agent
    controls a vehicle and is tasked with driving the vehicle from it's start
    location to a destination location while avoiding crashing into other vehicles.

    This environment requires each agent to navigate in the world while also
    taking care to avoid crashing into other agents. The dynamics and
    observations of the environment are such that avoiding collisions requires
    some planning in order for the vehicle to brake in time or maintain a good
    speed. Depending on the grid layout, the environment will require agents to
    reason about and possibly coordinate with the other vehicles.

    Possible Agents
    ---------------
    The environment supports two or more agents, depending on the grid. It is possible
    for some agents to finish the episode before other agents by either crashing or
    reaching their destination, and so not all agents are guaranteed to be active at
    the same time. All agents will be active at the start of the episode however.

    State Space
    -----------
    Each state is made up of the state of each vehicle, which in turn is defined by:

    - the `(x, y)` coordinates (x=column, y=row, with origin at the top-left square of
        the grid) of the vehicle,
    - the direction the vehicle is facing `NORTH=0`, `EAST=1`, `SOUTH=2`, `WEST=3`,
    - the speed of the vehicle: `REVERSE=0`, `STOPPED=1`, `FORWARD_SLOW=2`,
        `FORWARD_FAST=2`,
    - the `(x, y)` coordinate of the vehicles destination
    - whether the vehicle has reached it's destination or not: `1` or `0`
    - whether the vehicle has crashed or not: `1` or `0`
    - the minimum distance to the destination achieved by the vehicle in the current
        episode.
    - the initial distance of the vehicle to the destination at the start of the
        episode.

    Action Space
    ------------
    Each agent has 5 actions: `DO_NOTHING=0`, `ACCELERATE=1`, `DECELERATE=2`,
    `TURN_RIGHT=3`, `TURN_LEFT=4`

    Observation Space
    -----------------
    Each agent observes the cells in their local area, as well as their current speed,
    current location, their destination location, whether they've reached their
    destination, and whether they've crashed. The size of the local area observed is
    controlled by the `obs_dims` parameter (default = `(3, 1, 1)`, 3 cells in front,
    one cell behind, and 1 cell each side, giving a observation size of 5x3).For each
    cell in the observed area the agent observes whether the cell contains a
    `VEHICLE=0`, `WALL=1`, `EMPTY=2`, or it's `DESTINATION=3`.

    All together each agent's observation is tuple of the form:

        ((local obs), speed, coord, destination coord, destination reached, crashed)

    Rewards
    -------
    If an agent crashes into a vehicle (or is crashed into) they receive a penalty of
    `-1.0`. A reward of `0.5` is given if the agent reaches it's destination.
    Additionally, agents receive a small reward each step they makes progress towards
    their destination (i.e. the agent reduces it's minimum distance achieved to the
    destination for the episode). The total amount of reward and agent receives for
    making progress is `0.5`, and is distributed evenly across all steps the agent
    makes progress. This means if the agent reaches their destination they will receive
    a total reward of `1.0` (`0.5` for reaching their destination, and `0.5` for
    progress).

    Dynamics
    --------
    Actions are deterministic and movement is determined by direction the vehicle is
    facing and it's speed:

    - Speed=0 (REVERSE) - vehicle moves one cell in the opposite direction to which it
        is facing (vehicles cannot turn while in reverse)
    - Speed=1 (STOPPED) - vehicle remains in same cell
    - Speed=2 (FORWARD_SLOW) - vehicle move one cell in facing direction
    - Speed=3 (FORWARD_FAST) - vehicle moves two cells in facing direction

    Accelerating increases speed by 1, while deceleration decreased speed by 1. If the
    vehicle will hit a wall or another vehicle when moving from one cell to another then
    it remains in it's current cell and it's speed is reduced to 1 (STOPPED).

    Starting State
    --------------
    Each agent is randomly assigned to one of the possible starting locations on the
    grid and one of the possible destination locations, with no two agents starting in
    the same location or having the same destination location. The possible start and
    destination locations are determined by the grid layout being used.

    Episodes End
    ------------
    Episodes end when all agents have either reached their destination or crashed. By
    default a `max_episode_steps` is also set for each Driving environment. The default
    value is `50` steps, but this may need to be adjusted when using larger grids (this
    can be done by manually specifying a value for `max_episode_steps` when creating the
    environment with `posggym.make`).

    Arguments
    ---------

    - `grid` - the grid layout to use. This can either be a string specifying one of
        the supported grids, or a custom :class:`DrivingGrid` object
        (default = `"14x14RoundAbout"`).
    - `num_agents` - the number of agents in the environment (default = `2`).
    - `obs_dim` - the local observation dimensions, specifying how many cells in front,
        behind, and to each side the agent observes (default = `(3, 1, 1)`, resulting
        in the agent observing a 5x3 area: 3 in front, 1 behind, 1 to each side.)

    Available variants
    ------------------

    The Driving environment comes with a number of pre-built grid layouts which can be
    passed as an argument to `posggym.make`, to create different grids:

    | Grid name         | Max number of agents | Grid size |
    |-------------------|----------------------|---------- |
    | `3x3`             | 2                    | 3x3       |
    | `6x6Intersection` | 4                    | 6x6       |
    | `7x7Blocks`       | 4                    | 7x7       |
    | `7x7CrissCross`   | 6                    | 7x7       |
    | `7x7RoundAbout`   | 4                    | 7x7       |
    | `14x14Blocks`     | 4                    | 14x14     |
    | `14x14CrissCross` | 8                    | 14x14     |
    | `14x14RoundAbout` | 4                    | 14x14     |


    For example to use the Driving environment with the `7x7RoundAbout` grid and 2
    agents, you would use:

    ```python
    import posggym
    env = posggym.make('Driving-v1', grid="7x7RoundAbout", num_agents=2)
    ```

    Version History
    ---------------
    - `v1`: Major update:
        - Added agent's current location to observation space (to allow for the
          creation of heuristic policies for the environment and mimics GPS),
        - made it so vehicles speed is reduced to 0 if they crash or hit a wall (instead
          of remaining at their current speed),
        - removed obstacle collisions option entirely (since it wasn't really used, and
          not core to what the environment is testing),
        - updated reward so max return is 1.0 (0.5 for reaching destination, and 0.5
          from progress) and min return is -1.0 (-1.0 for crashing),
    - `v0`: Initial version

    References
    ----------
    - Adam Lerer and Alexander Peysakhovich. 2019. Learning Existing Social Conventions
    via Observationally Augmented Self-Play. In Proceedings of the 2019 AAAI/ACM
    Conference on AI, Ethics, and Society. 107–114.
    - Kevin R. McKee, Joel Z. Leibo, Charlie Beattie, and Richard Everett. 2022.
    Quantifying the Effects of Environment and Population Diversity in Multi-Agent
    Reinforcement Learning. Autonomous Agents and Multi-Agent Systems 36, 1 (2022), 1–16

    """

    metadata = {
        "render_modes": ["human", "ansi", "rgb_array", "rgb_array_dict"],
        "render_fps": 15,
    }

    def __init__(
        self,
        grid: Union[str, "DrivingGrid"] = "14x14RoundAbout",
        num_agents: int = 2,
        obs_dim: Tuple[int, int, int] = (3, 1, 1),
        render_mode: Optional[str] = None,
    ):
        super().__init__(
            DrivingModel(grid, num_agents, obs_dim),
            render_mode=render_mode,
        )
        self._obs_dim = obs_dim
        self.renderer = None
        self._agent_imgs = None

    def render(self):
        if self.render_mode is None:
            assert self.spec is not None
            logger.warn(
                "You are calling render method without specifying any render mode. "
                "You can specify the render_mode at initialization, "
                f'e.g. posggym.make("{self.spec.id}", render_mode="rgb_array")'
            )
            return
        if self.render_mode == "ansi":
            return self._render_ansi()
        return self._render_img()

    def _render_ansi(self):
        model: DrivingModel = self.model  # type: ignore
        grid_str = model.grid.get_ascii_repr(
            [vs.coord for vs in self._state],
            [vs.facing_dir for vs in self._state],
            [vs.dest_coord for vs in self._state],
        )

        output = [
            f"Step: {self._step_num}",
            grid_str,
        ]
        if self._last_actions is not None:
            action_str = ", ".join(
                [ACTIONS_STR[a] for a in self._last_actions.values()]
            )
            output.insert(1, f"Actions: <{action_str}>")
            output.append(f"Rewards: <{self._last_rewards}>")

        return "\n".join(output) + "\n"

    def _render_img(self):
        model: DrivingModel = self.model  # type: ignore

        import posggym.envs.grid_world.render as render_lib

        if self.renderer is None:
            self.renderer = render_lib.GWRenderer(
                self.render_mode,
                model.grid,
                render_fps=self.metadata["render_fps"],
                env_name="Driving",
            )

        if self._agent_imgs is None:
            self._agent_imgs = {
                i: render_lib.GWTriangle(
                    (0, 0),
                    self.renderer.cell_size,
                    render_lib.get_agent_color(i)[0],
                    Direction.NORTH,
                )
                for i in self.possible_agents
            }

        observed_coords = []
        for vs in self._state:
            observed_coords.extend(model.get_obs_coords(vs.coord, vs.facing_dir))

        render_objects = []
        # Add agent destination locations
        for i, vs in enumerate(self._state):
            # use alternative agent color so dest squares slightly different to vehicle
            # color
            render_objects.append(
                render_lib.GWRectangle(
                    vs.dest_coord,
                    self.renderer.cell_size,
                    render_lib.get_agent_color(i)[1],
                )
            )

        # Add agents
        for i, vs in enumerate(self._state):
            agent_obj = self._agent_imgs[str(i)]
            agent_obj.coord = vs.coord
            agent_obj.facing_dir = vs.facing_dir
            render_objects.append(agent_obj)

        agent_coords_and_dirs = {
            str(i): (vs.coord, vs.facing_dir) for i, vs in enumerate(self._state)
        }

        # Add visualization for crashed agents
        for i, vs in enumerate(self._state):
            if vs.crashed:
                render_objects.append(
                    render_lib.GWCircle(
                        vs.coord,
                        self.renderer.cell_size,
                        render_lib.get_color("yellow"),
                    )
                )

        if self.render_mode in ("human", "rgb_array"):
            return self.renderer.render(render_objects, observed_coords)
        return self.renderer.render_agents(
            render_objects,
            agent_coords_and_dirs,
            agent_obs_dims=(*self._obs_dim, self._obs_dim[-1]),
            observed_coords=observed_coords,
        )

    def close(self) -> None:
        if self.renderer is not None:
            self.renderer.close()
            self.renderer = None


class DrivingModel(M.POSGModel[DState, DObs, DAction]):
    """Driving Problem Model.

    Parameters
    ----------
    grid : DrivingGrid
        the grid environment for the model scenario
    num_agents : int
        the number of agents in the model scenario
    obs_dims : (int, int, int)
        number of cells in front, behind, and to the side that each agent
        can observe
    """

    R_CRASH_VEHICLE = -1.0
    R_DESTINATION_REACHED = 0.5
    R_PROGRESS_TOTAL = 0.5

    def __init__(
        self,
        grid: Union[str, "DrivingGrid"],
        num_agents: int,
        obs_dim: Tuple[int, int, int],
    ):
        if isinstance(grid, str):
            assert grid in SUPPORTED_GRIDS, (
                f"Unsupported grid '{grid}'. If grid argument is a string it must be "
                f"one of: {SUPPORTED_GRIDS.keys()}."
            )
            grid_info = SUPPORTED_GRIDS[grid]
            supported_num_agents: int = grid_info["supported_num_agents"]
            assert 0 < num_agents <= supported_num_agents, (
                f"Driving grid `{grid}` does not support {num_agents} agents. The "
                f"supported number of agents is from 1 up to {supported_num_agents}."
            )
            grid = parse_grid_str(
                grid_info["grid_str"], grid_info["supported_num_agents"]
            )
        else:
            assert 0 < num_agents <= grid.supported_num_agents, (
                f"Supplied DrivingGrid `{grid}` does not support {num_agents} agents. "
                "The supported number of agents is from 1 up to "
                f"{grid.supported_num_agents}."
            )

        assert obs_dim[0] > 0 and obs_dim[1] >= 0 and obs_dim[2] >= 0
        self._grid = grid
        self.obs_dim = obs_dim
        self._obs_front, self._obs_back, self._obs_side = obs_dim

        def _coord_space():
            return spaces.Tuple(
                (spaces.Discrete(self.grid.width), spaces.Discrete(self.grid.height))
            )

        self.possible_agents = tuple(str(i) for i in range(num_agents))
        self.state_space = spaces.Tuple(
            tuple(
                spaces.Tuple(
                    (
                        _coord_space(),
                        spaces.Discrete(len(Direction)),
                        spaces.Discrete(len(Speed)),
                        _coord_space(),  # destination coord
                        spaces.Discrete(2),  # destination reached
                        spaces.Discrete(2),  # crashed
                        # min and init distance to destination
                        # set these to upper bound of min shortest path distance, so
                        # state space works for generated grids as well
                        spaces.Discrete(self.grid.width * self.grid.height),
                        spaces.Discrete(self.grid.width * self.grid.height),
                    )
                )
                for _ in range(len(self.possible_agents))
            )
        )
        self.action_spaces = {
            i: spaces.Discrete(len(ACTIONS)) for i in self.possible_agents
        }

        obs_depth = self._obs_front + self._obs_back + 1
        obs_width = (2 * self._obs_side) + 1
        self.observation_spaces = {
            i: spaces.Tuple(
                (
                    spaces.Tuple(
                        tuple(
                            spaces.Discrete(len(CELL_OBS))
                            for _ in range(obs_depth * obs_width)
                        )
                    ),
                    spaces.Discrete(len(Speed)),
                    _coord_space(),  # current coord
                    _coord_space(),  # dest coord,
                    spaces.Discrete(2),  # dest reached
                    spaces.Discrete(2),  # crashed
                )
            )
            for i in self.possible_agents
        }
        self.is_symmetric = True

    @property
    def reward_ranges(self) -> Dict[str, Tuple[float, float]]:
        return {
            i: (
                self.R_CRASH_VEHICLE,
                self.R_DESTINATION_REACHED + self.R_PROGRESS_TOTAL,
            )
            for i in self.possible_agents
        }

    @property
    def rng(self) -> seeding.RNG:
        if self._rng is None:
            self._rng, _ = seeding.std_random()
        return self._rng

    def get_agents(self, state: DState) -> List[str]:
        return list(self.possible_agents)

    @property
    def grid(self) -> "DrivingGrid":
        """The underlying grid for this model instance."""
        return self._grid

    @grid.setter
    def grid(self, grid: "DrivingGrid"):
        assert (self._grid.height, self._grid.width) == (grid.height, grid.width)
        self._grid = grid

    def sample_initial_state(self) -> DState:
        state = []
        chosen_start_coords: Set[Coord] = set()
        chosen_dest_coords: Set[Coord] = set()
        for i in range(len(self.possible_agents)):
            start_coords_i = self.grid.start_coords[i]
            avail_start_coords = start_coords_i.difference(chosen_start_coords)
            start_coord = self.rng.choice(list(avail_start_coords))
            chosen_start_coords.add(start_coord)

            dest_coords_i = self.grid.dest_coords[i]
            avail_dest_coords = dest_coords_i.difference(chosen_dest_coords)
            if start_coord in avail_dest_coords:
                avail_dest_coords.remove(start_coord)
            dest_coord = self.rng.choice(list(avail_dest_coords))
            chosen_dest_coords.add(dest_coord)

            dest_dist = self.grid.get_shortest_path_distance(start_coord, dest_coord)

            state_i = VehicleState(
                coord=start_coord,
                facing_dir=INIT_DIR,
                speed=INIT_SPEED,
                dest_coord=dest_coord,
                dest_reached=int(False),
                crashed=int(False),
                min_dest_dist=dest_dist,
                init_dest_dist=dest_dist,
            )
            state.append(state_i)
        return tuple(state)

    def sample_agent_initial_state(self, agent_id: str, obs: DObs) -> DState:
        agent_idx = int(agent_id)
        agent_start_coord = obs[2]
        agent_dest_coord = obs[3]

        state = []
        chosen_start_coords: Set[Coord] = set()
        chosen_dest_coords: Set[Coord] = set()

        chosen_start_coords.add(agent_start_coord)
        chosen_dest_coords.add(agent_dest_coord)

        for i in range(len(self.possible_agents)):
            if i == agent_idx:
                start_coord = agent_start_coord
            else:
                start_coords_i = self.grid.start_coords[i]
                avail_coords = start_coords_i.difference(chosen_start_coords)
                start_coord = self.rng.choice(list(avail_coords))
                chosen_start_coords.add(start_coord)

            if i == agent_idx:
                dest_coord = agent_dest_coord
            else:
                dest_coords_i = self.grid.dest_coords[i]
                avail_coords = dest_coords_i.difference(chosen_dest_coords)
                if start_coord in avail_coords:
                    avail_coords.remove(start_coord)
                dest_coord = self.rng.choice(list(avail_coords))
                chosen_dest_coords.add(dest_coord)

            dest_dist = self.grid.get_shortest_path_distance(start_coord, dest_coord)

            state_i = VehicleState(
                coord=start_coord,
                facing_dir=INIT_DIR,
                speed=INIT_SPEED,
                dest_coord=dest_coord,
                dest_reached=int(False),
                crashed=int(False),
                min_dest_dist=dest_dist,
                init_dest_dist=dest_dist,
            )
            state.append(state_i)
        return tuple(state)

    def sample_initial_obs(self, state: DState) -> Dict[str, DObs]:
        return self._get_obs(state)

    def step(
        self, state: DState, actions: Dict[str, DAction]
    ) -> M.JointTimestep[DState, DObs]:
        assert all(a_i in ACTIONS for a_i in actions.values())
        next_state = self._get_next_state(state, actions)
        obs = self._get_obs(next_state)
        rewards = self._get_rewards(state, next_state)
        terminated = {
            i: bool(next_state[int(i)].dest_reached or next_state[int(i)].crashed)
            for i in self.possible_agents
        }
        truncated = {i: False for i in self.possible_agents}
        all_done = all(terminated.values())

        info: Dict[str, Dict] = {i: {} for i in self.possible_agents}
        for idx in range(len(self.possible_agents)):
            if next_state[idx].dest_reached:
                outcome_i = M.Outcome.WIN
            elif next_state[idx].crashed:
                outcome_i = M.Outcome.LOSS
            else:
                outcome_i = M.Outcome.NA
            info[str(idx)]["outcome"] = outcome_i

        return M.JointTimestep(
            next_state, obs, rewards, terminated, truncated, all_done, info
        )

    def _get_next_state(
        self, state: DState, actions: Dict[str, DAction]
    ) -> Tuple[DState, List[bool]]:
        exec_order = list(range(len(self.possible_agents)))
        self.rng.shuffle(exec_order)

        next_state = list(state)
        vehicle_coords = {vs.coord: idx for idx, vs in enumerate(state)}
        for idx in exec_order:
            state_i = state[idx]
            next_state_i = next_state[idx]
            if state_i.dest_reached or state_i.crashed or next_state_i.crashed:
                # already in terminal/rewarded state, or was crashed into this step
                continue

            action_i = actions[str(idx)]

            vehicle_coords.pop(state_i.coord)

            next_speed = self.get_next_speed(action_i, state_i.speed)
            move_dir = self.get_move_direction(action_i, next_speed, state_i.facing_dir)
            next_dir = self.get_next_direction(action_i, next_speed, state_i.facing_dir)
            next_coord, crashed, hit_vehicle = self._get_next_coord(
                state_i.coord, next_speed, move_dir, vehicle_coords
            )
            if next_coord == state_i.coord:
                # crashed or hit a wall
                next_speed = Speed.STOPPED

            min_dest_dist = min(
                state_i.min_dest_dist,
                self.grid.get_shortest_path_distance(next_coord, state_i.dest_coord),
            )

            if crashed:
                # update state of vehicle that was hit
                jdx = vehicle_coords[hit_vehicle]
                next_state_j = next_state[jdx]
                next_state[jdx] = VehicleState(
                    coord=next_state_j.coord,
                    facing_dir=next_state_j.facing_dir,
                    speed=next_state_j.speed,
                    dest_coord=next_state_j.dest_coord,
                    dest_reached=next_state_j.dest_reached,
                    crashed=int(True),
                    min_dest_dist=next_state_j.min_dest_dist,
                    init_dest_dist=next_state_j.init_dest_dist,
                )

            next_state[idx] = VehicleState(
                coord=next_coord,
                facing_dir=next_dir,
                speed=next_speed,
                dest_coord=state_i.dest_coord,
                dest_reached=int(next_coord == state_i.dest_coord),
                crashed=int(crashed),
                min_dest_dist=min_dest_dist,
                init_dest_dist=state_i.init_dest_dist,
            )

            vehicle_coords[next_coord] = idx

        return tuple(next_state)

    @staticmethod
    def get_move_direction(
        action: DAction, speed: Speed, curr_dir: Direction
    ) -> Direction:
        if speed == Speed.REVERSE:
            # No turning while in reverse,
            # so movement dir is always just the opposite of current direction
            return Direction((curr_dir + 2) % len(Direction))
        return DrivingModel.get_next_direction(action, speed, curr_dir)

    @staticmethod
    def get_next_direction(
        action: DAction, speed: Speed, curr_dir: Direction
    ) -> Direction:
        if action == TURN_RIGHT and speed != Speed.REVERSE:
            return Direction((curr_dir + 1) % len(Direction))
        if action == TURN_LEFT and speed != Speed.REVERSE:
            return Direction((curr_dir - 1) % len(Direction))
        return curr_dir

    @staticmethod
    def get_next_speed(action: DAction, curr_speed: Speed) -> Speed:
        if action == DO_NOTHING:
            return curr_speed
        if action in (TURN_LEFT, TURN_RIGHT):
            if curr_speed == Speed.FORWARD_FAST:
                return Speed.FORWARD_SLOW
            return curr_speed
        if action == ACCELERATE:
            return Speed(min(curr_speed + 1, Speed.FORWARD_FAST))
        return Speed(max(curr_speed - 1, Speed.REVERSE))

    def _get_next_coord(
        self,
        curr_coord: Coord,
        speed: Speed,
        move_dir: Direction,
        vehicle_coords: Set[Coord],
    ) -> Tuple[Coord, bool, Optional[Coord]]:
        # assumes curr_coord isn't in vehicle coords
        next_coord = curr_coord
        crashed = False
        hit_vehicle_coord = None
        for _ in range(abs(speed - Speed.STOPPED)):
            next_coord = self.grid.get_next_coord(
                curr_coord, move_dir, ignore_blocks=False
            )
            if next_coord in vehicle_coords:
                crashed = True
                hit_vehicle_coord = next_coord
                next_coord = curr_coord
                break
            curr_coord = next_coord

        return (next_coord, crashed, hit_vehicle_coord)

    def _get_obs(self, state: DState) -> Dict[str, DObs]:
        obs: Dict[str, DObs] = {}
        for i in self.possible_agents:
            idx = int(i)
            local_cell_obs = self._get_local_cell__obs(
                idx,
                [vs.coord for vs in state],
                state[idx].facing_dir,
                state[idx].dest_coord,
            )
            obs[i] = (
                local_cell_obs,
                state[idx].speed,
                state[idx].coord,
                state[idx].dest_coord,
                state[idx].dest_reached,
                state[idx].crashed,
            )
        return obs

    def _get_local_cell__obs(
        self,
        agent_idx: int,
        vehicle_coords: Sequence[Coord],
        facing_dir: Direction,
        dest_coord: Coord,
    ) -> Tuple[int, ...]:
        obs_depth = self._obs_front + self._obs_back + 1
        obs_width = (2 * self._obs_side) + 1
        agent_coord = vehicle_coords[agent_idx]

        cell_obs = []
        for row, col in product(range(obs_depth), range(obs_width)):
            obs_grid_coord = self._map_obs_to_grid_coord(
                (col, row), agent_coord, facing_dir
            )
            if obs_grid_coord is None or obs_grid_coord in self.grid.block_coords:
                cell_obs.append(WALL)
            elif obs_grid_coord in vehicle_coords:
                cell_obs.append(VEHICLE)
            elif obs_grid_coord == dest_coord:
                cell_obs.append(DESTINATION)
            else:
                cell_obs.append(EMPTY)
        return tuple(cell_obs)

    def _map_obs_to_grid_coord(
        self, obs_coord: Coord, agent_coord: Coord, facing_dir: Direction
    ) -> Optional[Coord]:
        if facing_dir == Direction.NORTH:
            grid_row = agent_coord[1] + obs_coord[1] - self._obs_front
            grid_col = agent_coord[0] + obs_coord[0] - self._obs_side
        elif facing_dir == Direction.EAST:
            grid_row = agent_coord[1] + obs_coord[0] - self._obs_side
            grid_col = agent_coord[0] - obs_coord[1] + self._obs_front
        elif facing_dir == Direction.SOUTH:
            grid_row = agent_coord[1] - obs_coord[1] + self._obs_front
            grid_col = agent_coord[0] - obs_coord[0] + self._obs_side
        else:
            grid_row = agent_coord[1] - obs_coord[0] + self._obs_side
            grid_col = agent_coord[0] + obs_coord[1] - self._obs_front

        if 0 <= grid_row < self.grid.height and 0 <= grid_col < self.grid.width:
            return (grid_col, grid_row)
        return None

    def get_obs_coords(self, origin: Coord, facing_dir: Direction) -> List[Coord]:
        """Get the list of coords observed by agent at origin."""
        obs_depth = self._obs_front + self._obs_back + 1
        obs_width = (2 * self._obs_side) + 1
        obs_coords: List[Coord] = []
        for col, row in product(range(obs_width), range(obs_depth)):
            obs_grid_coord = self._map_obs_to_grid_coord((col, row), origin, facing_dir)
            if obs_grid_coord is not None:
                obs_coords.append(obs_grid_coord)
        return obs_coords

    def _get_rewards(self, state: DState, next_state: DState) -> Dict[str, float]:
        rewards: Dict[str, float] = {}
        for i in self.possible_agents:
            idx = int(i)
            if state[idx].crashed or state[idx].dest_reached:
                # already in terminal/rewarded state
                r_i = 0.0
            elif next_state[idx].crashed:
                # crashed into a vehicle this step
                r_i = self.R_CRASH_VEHICLE
            elif next_state[idx].dest_reached:
                r_i = self.R_DESTINATION_REACHED
            else:
                r_i = 0.0

            progress = state[idx].min_dest_dist - next_state[idx].min_dest_dist
            r_i += self.R_PROGRESS_TOTAL * max(0, progress / state[idx].init_dest_dist)

            rewards[i] = r_i
        return rewards


class DrivingGrid(Grid):
    """A grid for the Driving Problem."""

    def __init__(
        self,
        grid_width: int,
        grid_height: int,
        block_coords: Set[Coord],
        start_coords: List[Set[Coord]],
        dest_coords: List[Set[Coord]],
    ):
        super().__init__(grid_width, grid_height, block_coords)
        assert len(start_coords) == len(dest_coords)
        self.start_coords = start_coords
        self.dest_coords = dest_coords
        self.shortest_paths = self.get_all_shortest_paths(set.union(*dest_coords))

    @property
    def supported_num_agents(self) -> int:
        """Get the number of agents supported by this grid."""
        return len(self.start_coords)

    def get_shortest_path_distance(self, coord: Coord, dest: Coord) -> int:
        """Get the shortest path distance from coord to destination."""
        return int(self.shortest_paths[dest][coord])

    def get_max_shortest_path_distance(self) -> int:
        """Get the longest shortest path distance to any destination."""
        return int(max([max(d.values()) for d in self.shortest_paths.values()]))

    def get_ascii_repr(
        self,
        vehicle_coords: List[Coord],
        vehicle_dirs: List[Direction],
        vehicle_dests: List[Coord],
    ) -> str:
        """Get ascii repr of grid."""
        grid_repr = []
        for row in range(self.height):
            row_repr = []
            for col in range(self.width):
                coord = (col, row)
                if coord in self.block_coords:
                    row_repr.append("#")
                elif coord in vehicle_dests:
                    row_repr.append("D")
                else:
                    row_repr.append(".")
            grid_repr.append(row_repr)

        for coord, direction in zip(vehicle_coords, vehicle_dirs):
            grid_repr[coord[0]][coord[1]] = DIRECTION_ASCII_REPR[direction]

        return "\n".join([" ".join(r) for r in grid_repr])


def parse_grid_str(grid_str: str, supported_num_agents: int) -> DrivingGrid:
    """Parse a str representation of a grid.

    Notes on grid str representation:

    . = empty/unallocated cell
    # = a block
    0, 1, ..., 9 = starting location for agent with given index
    + = starting point for any agent
    a, b, ..., j = destination location for agent with given index
                   (a=0, b=1, ..., j=9)
    - = destination location for any agent

    Examples (" " quotes and newline chars omitted):

    1. A 3x3 grid with two agents, one block, and where each agent has a single
    starting location and a single destination location.

    a1.
    .#.
    .0.

    2. A 6x6 grid with 4 common start and destination locations and many
       blocks. This grid can support up to 4 agents.

    +.##+#
    ..+#.+
    #.###.
    #.....
    -..##.
    #-.-.-

    """
    row_strs = grid_str.splitlines()
    assert len(row_strs) > 1
    assert all(len(row) == len(row_strs[0]) for row in row_strs)
    assert len(row_strs[0]) > 1

    grid_height = len(row_strs)
    grid_width = len(row_strs[0])

    agent_start_chars = set(["+"] + [str(i) for i in range(10)])
    agent_dest_chars = set(["-"] + list("abcdefghij"))

    block_coords: Set[Coord] = set()
    shared_start_coords: Set[Coord] = set()
    agent_start_coords_map: Dict[int, Set[Coord]] = {}
    shared_dest_coords: Set[Coord] = set()
    agent_dest_coords_map: Dict[int, Set[Coord]] = {}
    for r, c in product(range(grid_height), range(grid_width)):
        coord = (c, r)
        char = row_strs[r][c]

        if char == "#":
            block_coords.add(coord)
        elif char in agent_start_chars:
            if char != "+":
                agent_id = int(char)
                if agent_id not in agent_start_coords_map:
                    agent_start_coords_map[agent_id] = set()
                agent_start_coords_map[agent_id].add(coord)
            else:
                shared_start_coords.add(coord)
        elif char in agent_dest_chars:
            if char != "-":
                agent_id = ord(char) - ord("a")
                if agent_id not in agent_dest_coords_map:
                    agent_dest_coords_map[agent_id] = set()
                agent_dest_coords_map[agent_id].add(coord)
            else:
                shared_dest_coords.add(coord)

    assert (
        len(shared_start_coords) + len(agent_start_coords_map) >= supported_num_agents
    )
    assert len(shared_dest_coords) + len(agent_dest_coords_map) >= supported_num_agents

    included_agent_ids = list({*agent_start_coords_map, *agent_dest_coords_map})
    if len(included_agent_ids) > 0:
        assert max(included_agent_ids) < supported_num_agents

    start_coords: List[Set[Coord]] = []
    dest_coords: List[Set[Coord]] = []
    for i in range(supported_num_agents):
        agent_start_coords = set(shared_start_coords)
        agent_start_coords.update(agent_start_coords_map.get(i, {}))
        start_coords.append(agent_start_coords)

        agent_dest_coords = set(shared_dest_coords)
        agent_dest_coords.update(agent_dest_coords_map.get(i, {}))
        dest_coords.append(agent_dest_coords)

    return DrivingGrid(
        grid_width=grid_width,
        grid_height=grid_height,
        block_coords=block_coords,
        start_coords=start_coords,
        dest_coords=dest_coords,
    )


#  (grid_make_fn, max step_limit, )
SUPPORTED_GRIDS: Dict[str, Dict[str, Any]] = {
    "3x3": {
        "grid_str": ("a1.\n" ".#.\n" ".0b\n"),
        "supported_num_agents": 2,
        "max_episode_steps": 15,
    },
    "6x6Intersection": {
        "grid_str": ("##0b##\n" "##..##\n" "d....3\n" "2....c\n" "##..##\n" "##a1##\n"),
        "supported_num_agents": 4,
        "max_episode_steps": 20,
    },
    "7x7Blocks": {
        "grid_str": (
            "#-...-#\n"
            "-##.##+\n"
            ".##.##.\n"
            ".......\n"
            ".##.##.\n"
            "-##.##+\n"
            "#+...+#\n"
        ),
        "supported_num_agents": 4,
        "max_episode_steps": 50,
    },
    "7x7CrissCross": {
        "grid_str": (
            "#-#-#-#\n"
            "-.....+\n"
            "#.#.#.#\n"
            "-.....+\n"
            "#.#.#.#\n"
            "-.....+\n"
            "#+#+#+#\n"
        ),
        "supported_num_agents": 6,
        "max_episode_steps": 50,
    },
    "7x7RoundAbout": {
        "grid_str": (
            "#-...-#\n"
            "-##.##+\n"
            ".#...#.\n"
            "...#...\n"
            ".#...#.\n"
            "-##.##+\n"
            "#+...+#\n"
        ),
        "supported_num_agents": 4,
        "max_episode_steps": 50,
    },
    "14x14Blocks": {
        "grid_str": (
            "#-..........-#\n"
            "-###.####.###+\n"
            ".###.####.###.\n"
            ".###.####.###.\n"
            "..............\n"
            ".###.####.###.\n"
            ".###.####.###.\n"
            ".###.####.###.\n"
            ".###.####.###.\n"
            "..............\n"
            ".###.####.###.\n"
            ".###.####.###.\n"
            "-###.####.###+\n"
            "#+..........+#\n"
        ),
        "supported_num_agents": 4,
        "max_episode_steps": 50,
    },
    "14x14CrissCross": {
        "grid_str": (
            "##-##-##-##-##\n"
            "##.##.##.##.##\n"
            "-............+\n"
            "##.##.##.##.##\n"
            "##.##.##.##.##\n"
            "-............+\n"
            "##.##.##.##.##\n"
            "##.##.##.##.##\n"
            "-............+\n"
            "##.##.##.##.##\n"
            "##.##.##.##.##\n"
            "-............+\n"
            "##.##.##.##.##\n"
            "##+##+##+##+##\n"
        ),
        "supported_num_agents": 8,
        "max_episode_steps": 50,
    },
    "14x14RoundAbout": {
        "grid_str": (
            "#-..........-#\n"
            "-#####..#####+\n"
            ".#####..#####.\n"
            ".#####..#####.\n"
            ".###......###.\n"
            ".###......###.\n"
            "......##......\n"
            "......##......\n"
            ".###......###.\n"
            ".###......###.\n"
            ".#####..#####.\n"
            ".#####..#####.\n"
            "-#####..#####+\n"
            "#+..........+#\n"
        ),
        "supported_num_agents": 4,
        "max_episode_steps": 50,
    },
}



"""Shortest path policy for Driving Gen envs."""
from __future__ import annotations

from typing import TYPE_CHECKING

from posggym.agents.grid_world.driving.shortest_path import DrivingShortestPathPolicy


if TYPE_CHECKING:
    from posggym.agents.policy import PolicyID
    from posggym.model import POSGModel


class DrivingGenShortestPathPolicy(DrivingShortestPathPolicy):
    """Shortest Path Policy for the Driving Gen environment.

    This policy sets the preferred action as the one which is on the shortest
    path to the agent's goal and which doesn't leave agent in same position. Note,
    the policy is stochastic and will sample from a distribution over the preferred
    actions, since there may be multiple actions which are on the shortest path.

    This policy is the same as the shortest path policy for the
    [Driving](/environments/grid_world/driving) environment, except that it recomputes
    the shortest paths when reset (at the start of each episode), since the environment
    grid layout is regenerated at the start of each episode. Additionally, it only
    computes the shortest path for the agent's current destination, rather than for
    all destinations.

    Arguments
    ---------
    aggressiveness : float
        The aggressiveness of the policy towards other vehicles. A value of 0.0 means
        the policy will always stop when it sees another vehicle, while a value of 1.0
        means the policy will ignore other vehicles and always take the shortest path
        action. Values in between will change how far away another vehicle needs to
        be before the policy will stop. Default is 1.0.

    """

    def __init__(
        self,
        model: POSGModel,
        agent_id: str,
        policy_id: PolicyID,
        aggressiveness: float = 1.0,
    ):
        super().__init__(
            model,
            agent_id,
            policy_id,
            aggressiveness=aggressiveness,
            precompute_shortest_paths=False,
        )

    def reset(self, *, seed: int | None = None):
        super().reset(seed=seed)
        self._grid = self.model.grid
        self.shortest_paths = {}
